# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import  # noqa

from functools import wraps

from psycopg import errors

from restalchemy.storage.sql.dialect import base
from restalchemy.storage.sql.dialect import exceptions as exc

SAVEPOINT_EXP_MAP = {
    "savepoint": "SAVEPOINT {name};",
    "rollback": "ROLLBACK TO SAVEPOINT {name};",
    "release": "RELEASE SAVEPOINT {name};",
}


def handle_database_errors(func):
    """
    A decorator to handle PostgreSQL database errors.

    This decorator catches specific database errors from the
    `psycopg` library and raises them as :class:`.DeadLock` or
    :class:`.Conflict` exceptions when appropriate.

    :param func: The function to be wrapped.
    :type func: callable
    :return: A function that wraps `func` and handles database errors.
    :rtype: callable
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except errors.DeadlockDetected as e:
            raise exc.DeadLock(code=e.sqlstate, message=str(e))
        except (errors.UniqueViolation, errors.ForeignKeyViolation) as e:
            raise exc.Conflict(code=e.sqlstate, message=str(e))

    return wrapper


class PgSQLInsert(base.BaseInsertCommand):
    EXPRESSION = 'INSERT INTO "%s" (%s) VALUES (%s)'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL insert command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLUpdate(base.BaseUpdateCommand):
    EXPRESSION = 'UPDATE "%s" SET %s WHERE %s'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL update command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLDelete(base.BaseDeleteCommand):
    EXPRESSION = 'DELETE FROM "%s" WHERE %s'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL delete command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLBatchDelete(base.BaseBatchDelete):
    EXPRESSION_IN = 'DELETE FROM "%s" WHERE %s = ANY(%s)'
    EXPRESSION_FILTER = 'DELETE FROM "%s" WHERE %s'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL batch delete command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLSelect(base.BaseSelectCommand):
    EXPRESSION = 'SELECT %s FROM "%s"'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL select command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLCustomSelect(base.BaseCustomSelectCommand):
    EXPRESSION = 'SELECT %s FROM "%s"'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL custom select command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSQLCount(base.BaseCountCommand):
    EXPRESSION = 'SELECT COUNT(*) as count FROM "%s"'

    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL count command.

        This method utilizes the base class's execute method to perform the
        command execution. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class PgSqlOrmDialectCommand(base.BaseOrmDialectCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the PostgreSQL ORM command.

        This method compiles the query using the `get_statement` method and
        retrieves the values required for the query execution using the
        `get_values` method. Subsequently, it executes the SQL command using
        the parent class's `execute` method, passing the compiled statement
        and values to it. It is decorated with `handle_database_errors` to
        handle any PostgreSQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the query execution.
        """

        return super().execute()


class PgSQLDialect(base.AbstractDialect):
    DIALECT_NAME = "postgresql"

    def orm_command(self, table, query, session):
        """
        Creates a new PostgreSQL ORM command for the given table, query, and
        session.

        This method creates an instance of `PgSqlOrmDialectCommand` and
        initializes it with the given table, query, and session.

        :param table: The table to be used in the ORM dialect command.
        :param query: The query object that contains the SQL query to be
            compiled.
        :param session: The session to be used for executing the command.
        :return: The created PostgreSQL ORM command.
        :rtype: PgSqlOrmDialectCommand
        """

        return PgSqlOrmDialectCommand(table, query, session=session)

    def insert(self, table, data, session):
        """
        Inserts data into the specified table using a PostgreSQL dialect command.

        This method creates an instance of `PgSQLInsert` to insert the
        provided data into the given table within the specified session
        context.

        :param table: The table into which data is to be inserted.
        :param data: The data to be inserted into the table.
        :param session: The session to be used for executing the insert
            command.
        :return: An instance of `PgSQLInsert` configured with the table, data,
            and session.
        :rtype: PgSQLInsert
        """
        return PgSQLInsert(table, data, session=session)

    def update(self, table, ids, data, session):
        """
        Updates records in the specified table using a PostgreSQL dialect
        command.

        This method creates an instance of `PgSQLUpdate` to update records
        identified by the provided IDs in the given table with the specified
        data within the session context.

        :param table: The table where records are to be updated.
        :param ids: The IDs of the records to be updated.
        :param data: The data to update the records with.
        :param session: The session to be used for executing the update
            command.
        :return: An instance of `PgSQLUpdate` configured with the table,
            IDs, data, and session.
        :rtype: PgSQLUpdate
        """

        return PgSQLUpdate(table, ids, data, session=session)

    def delete(self, table, ids, session):
        """
        Deletes records from the specified table using a PostgreSQL dialect
        command.

        This method creates an instance of `PgSQLDelete` to delete
        records identified by the provided IDs from the given table within
        the session context.

        :param table: The table from which records are to be deleted.
        :param ids: The IDs of the records to be deleted.
        :param session: The session to be used for executing the delete
            command.
        :return: An instance of `PgSQLDelete` configured with the table, IDs,
            and session.
        :rtype: PgSQLDelete
        """

        return PgSQLDelete(table, ids, session=session)

    def select(
        self,
        table,
        filters,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Retrieves records from the specified table using a PostgreSQL dialect
        command.

        This method creates an instance of `PgSQLSelect` to select records
        from the given table based on the provided filters, with optional
        limit, order by, and locking constraints, within the session context.

        :param table: The table from which records are to be retrieved.
        :param filters: The filters to be used for selecting the records.
        :param session: The session to be used for executing the select
            command.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int, optional
        :param order_by: The ordering criteria for the selected records.
        :param locked: Whether to lock the selected records for update.
            Defaults to False.
        :type locked: bool
        :return: An instance of `PgSQLSelect` configured with the table,
            filters, session, limit, order by, and locked parameters.
        :rtype: PgSQLSelect
        """
        return PgSQLSelect(
            table=table,
            session=session,
            filters=filters,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )

    def custom_select(
        self,
        table,
        session,
        where_conditions,
        where_values,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Retrieves records from the specified table using a custom PostgreSQL
        dialect command.

        This method creates an instance of `PgSQLCustomSelect` to select
        records from the given table based on the provided where conditions
        and values, with optional limit, order by, and locking constraints,
        within the session context.

        :param table: The table from which records are to be retrieved.
        :param session: The session to be used for executing the select
            command.
        :param where_conditions: The conditions for the WHERE clause in the
            SQL statement.
        :param where_values: The values for the WHERE clause in the SQL
            statement.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int, optional
        :param order_by: The ordering criteria for the selected records.
        :param locked: Whether to lock the selected records for update.
            Defaults to False.
        :type locked: bool
        :return: An instance of `PgSQLCustomSelect` configured with the table,
            where conditions, where values, session, limit, order by, and
            locked parameters.
        :rtype: PgSQLCustomSelect
        """
        return PgSQLCustomSelect(
            table=table,
            session=session,
            where_conditions=where_conditions,
            where_values=where_values,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )

    def count(self, table, filters, session):
        """
        Retrieves the count of records from the specified table based on the
        provided filters.

        This method creates an instance of `PgSQLCount` to count records
        from the given table using the specified session and filters.

        :param table: The table from which records are to be counted.
        :param filters: The filters to be used for counting the records.
        :param session: The session to be used for executing the count
            command.
        :return: An instance of `PgSQLCount` configured with the table,
            filters, and session parameters.
        :rtype: PgSQLCount
        """
        return PgSQLCount(
            table=table,
            session=session,
            filters=filters,
        )



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base


class CountModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "test_count"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


class WithDbMigrationsCountTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-count-migration-502944"
    __FIRST_MIGRATION__ = "test-count-migration-502944"

    def test_count(self):
        target_cnt = 4

        cnt = CountModel.objects.count()

        self.assertEqual(target_cnt, cnt)

    def test_count_filter_single_row(self):
        target_cnt = 1

        cnt = CountModel.objects.count(filters={"foo_field2": "value2"})

        self.assertEqual(target_cnt, cnt)

    def test_count_filter_many_rows(self):
        target_cnt = 2

        cnt = CountModel.objects.count(filters={"foo_field1": filters.GT(2)})

        self.assertEqual(target_cnt, cnt)



# Copyright 2022 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import constants as ra_const
from restalchemy.common import status

OPENAPI_SPECIFICATION_3_0_3 = "3.0.3"

API_VERSION_V1 = "v1"

ERROR_SCHEMA = {
    "title": "error",
    "description": "Error occurred while processing request",
    "type": "object",
    "required": [
        "status",
        "json",
    ],
    "properties": {
        "status": {
            "type": "integer",
            "description": "The HTTP status code",
            "minimum": 400,
            "exclusiveMaximum": True,
            "example": 503,
        },
        "json": {
            "type": "object",
            "description": "Detail error map",
            "required": [
                "code",
                "type",
                "message",
            ],
            "properties": {
                "code": {
                    "type": "integer",
                    "description": "The HTTP status code",
                    "minimum": 400,
                    "exclusiveMaximum": True,
                    "example": 503,
                },
                "type": {
                    "type": "string",
                    "description": "Error class name",
                    "example": "OSError",
                },
                "message": {
                    "type": "string",
                    "description": "Error message",
                    "example": "A human-readable explanation of problem",
                },
            },
        },
    },
    "example": {
        "code": 400,
        "json": {
            "code": 400,
            "type": "ValidationErrorException",
            "message": "Validation error occurred.",
        },
    },
}

ERROR_RESPONSE = {
    "description": "Error response.",
    "content": {
        ra_const.CONTENT_TYPE_APPLICATION_JSON: {
            "schema": {"$ref": "#/components/schemas/Error"}
        }
    },
}

DEFAULT_RESPONSE = {"$ref": "#/components/responses/Error"}

OPENAPI_DELETE_RESPONSE = {
    status.HTTP_204_NO_CONTENT: {
        "description": "Delete entity",
    },
    "default": DEFAULT_RESPONSE,
}

OPENAPI_FILTER_RESPONSE = {
    status.HTTP_200_OK: {
        "description": "Get nested urls",
        "content": {
            ra_const.CONTENT_TYPE_APPLICATION_JSON: {
                "schema": {"type": "array", "items": {"type": "string"}}
            }
        },
    },
    "default": DEFAULT_RESPONSE,
}

OPENAPI_DEFAULT_RESPONSE = {
    status.HTTP_200_OK: {
        "description": "Get nested urls",
        "content": {ra_const.CONTENT_TYPE_APPLICATION_JSON: {"schema": {}}},
    },
    "default": DEFAULT_RESPONSE,
}


def build_openapi_create_response(ref_name):
    return {
        status.HTTP_201_CREATED: {
            "description": ref_name,
            "content": {
                ra_const.CONTENT_TYPE_APPLICATION_JSON: {
                    "schema": {"$ref": "#/components/schemas/{}".format(ref_name)}
                }
            },
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_get_update_response(ref_name):
    return {
        status.HTTP_200_OK: {
            "description": ref_name,
            "content": {
                ra_const.CONTENT_TYPE_APPLICATION_JSON: {
                    "schema": {"$ref": "#/components/schemas/{}".format(ref_name)}
                }
            },
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_list_model_response(ref_name):
    return {
        status.HTTP_200_OK: {
            "description": ref_name,
            "content": {
                ra_const.CONTENT_TYPE_APPLICATION_JSON: {
                    "schema": {
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/{}".format(ref_name)},
                    }
                }
            },
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_delete_response(ref_name):
    return {
        status.HTTP_204_NO_CONTENT: {
            "description": ref_name,
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_object_response(
    properties,
    code=200,
    description="",
    content_type=ra_const.CONTENT_TYPE_APPLICATION_JSON,
):
    """

    properties - dict as needed in openapi
    https://swagger.io/docs/specification/describing-responses/

    Ex:
    {
    "id": {"type": "integer", "description": "The user ID"}
    "username": {"type": "string", "description": "The user name"}
    }

    """
    return {
        code: {
            "description": description,
            "content": {
                content_type: {"schema": {"type": "object", "properties": properties}}
            },
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_response_octet_stream(description="Returns binary file"):
    return {
        status.HTTP_200_OK: {
            "description": description,
            "content": {
                ra_const.CONTENT_TYPE_OCTET_STREAM: {
                    "schema": {"format": "binary", "type": "string"}
                }
            },
        },
    }


def build_openapi_user_response(code=status.HTTP_200_OK, description="", **kwargs):
    """

    properties - dict as needed in openapi
    https://swagger.io/docs/specification/describing-responses/

    Ex:
    {
    "id": {"type": "integer", "description": "The user ID"}
    "username": {"type": "string", "description": "The user name"}
    }

    """
    if not kwargs:
        raise ValueError("**kwargs are required.")
    return {
        code: {
            "description": description,
            "content": {ra_const.CONTENT_TYPE_APPLICATION_JSON: {"schema": kwargs}},
        },
        "default": DEFAULT_RESPONSE,
    }


def build_openapi_req_body(description, content_type, schema):
    return {
        "description": description,
        "required": True,
        "content": {content_type: {"schema": schema}},
    }


def build_openapi_req_body_multipart(description, properties):

    props = {}
    # See openapi types, examples:
    #  {"format": "binary", "type": "string"} - for files
    #  {"format": "uuid", "type": "string"} - for uuids
    for name, pval in properties.items():
        props[name] = pval

    return build_openapi_req_body(
        description=description,
        content_type=ra_const.CONTENT_TYPE_MULTIPART,
        schema={"properties": props},
    )


def build_openapi_json_req_body(model_name):
    return build_openapi_req_body(
        model_name,
        ra_const.CONTENT_TYPE_APPLICATION_JSON,
        {"$ref": f"#/components/schemas/{model_name}"},
    )


def build_openapi_parameter(
    name,
    description="",
    required=True,
    openapi_type="string",
    param_type="path",
    schema=None,
):
    param = {
        "name": name,
        "description": description,
        "in": param_type,
        "schema": schema or {"type": openapi_type},
    }
    if param_type == "path" and required is not None:
        param["required"] = required
    return param



#    Copyright 2021 George Melikov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import unittest

from restalchemy.dm import types_network


class RecordNameTestCase(unittest.TestCase):
    def setUp(self):
        super(RecordNameTestCase, self).setUp()
        self.test_instance = types_network.RecordName()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate("ns1.ra.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("ns1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("n_s1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("n-1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("restalchemy.com"))
        self.assertTrue(self.test_instance.validate("a.b.c.d.1.2.3"))
        self.assertTrue(self.test_instance.validate("qa-auto-dnsidxbs"))

    def test_from_simple_type(self):
        self.assertEqual(self.test_instance.from_simple_type(".x."), ".x")
        self.assertEqual(self.test_instance.from_simple_type(".x.x."), ".x.x")
        self.assertEqual(self.test_instance.from_simple_type("@"), "")

    def test_to_simple_type(self):
        self.assertEqual(self.test_instance.to_simple_type(""), "@")
        self.assertEqual(self.test_instance.to_simple_type("xxx"), "xxx")

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("a..b.s"))
        self.assertFalse(self.test_instance.validate(".a.b.s"))
        self.assertFalse(self.test_instance.validate("a.b.s.."))
        self.assertFalse(self.test_instance.validate("my.москва.рф"))
        self.assertFalse(self.test_instance.validate("москва.рф"))
        self.assertFalse(self.test_instance.validate("ee.ёёё.ЕЁ"))


class RecordNameWithWildcardTestCase(unittest.TestCase):
    def setUp(self):
        super(RecordNameWithWildcardTestCase, self).setUp()
        self.test_instance = types_network.RecordNameWithWildcard()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate("*.ns1.ra.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("ns1.ra.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("*.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("restalchemy.com"))

    def test_from_simple_type(self):
        self.assertEqual(self.test_instance.from_simple_type("*.x."), "*.x")
        self.assertEqual(self.test_instance.from_simple_type("*.x.x."), "*.x.x")

    def test_to_simple_type(self):
        self.assertEqual(self.test_instance.to_simple_type("*.xxx"), "*.xxx")

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("*.a..b.s"))
        self.assertFalse(self.test_instance.validate("a.*.b.s"))
        self.assertFalse(self.test_instance.validate(".a.b.s"))
        self.assertFalse(self.test_instance.validate("*a.b.s"))
        self.assertFalse(self.test_instance.validate("a.b.s*"))
        self.assertFalse(self.test_instance.validate("a.b*.s"))
        self.assertFalse(self.test_instance.validate("a.b.s..*"))


class SrvNameTest(unittest.TestCase):
    def setUp(self):
        self.srv_record = types_network.SrvName()

    def test_validate(self):
        self.assertTrue(self.srv_record.validate("_sip._tcp.ra.ru"))
        self.assertTrue(self.srv_record.validate("_sip._tcp.tk"))
        self.assertTrue(self.srv_record.validate("_sip._tcp"))

        self.assertFalse(self.srv_record.validate("sip._tcp.ra.ru"))
        self.assertFalse(self.srv_record.validate("_sip.tcp.ra.ru"))
        self.assertFalse(self.srv_record.validate("hcb.jyg"))


class HostnameTest(unittest.TestCase):
    def setUp(self):
        super(HostnameTest, self).setUp()
        self.fqdn = types_network.Hostname()
        self.fqdn_2level = types_network.Hostname(min_levels=2)
        self.fqdn_with_leading_underscore = types_network.Hostname(
            allow_leading_underscore=True
        )
        self.fqdn_with_middle_underscore = types_network.Hostname(
            allow_middle_underscore=True
        )

    def test_validate(self):
        data = [
            "first-level",
            "test-me.me",
            "fe.fe",
            "aa",
            "a.bc",
            "1.2.3.4.com",
            "xn--kxae4bafwg.xn--pxaix.gr",
            "a23456789-123456789-123456789-123456789-123456789-123456789-123.b23.com",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a234567.com",
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "com",
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn.validate(fqdn), fqdn)

    def test_validate_underscore(self):
        data = ["_acme-challenge.example.com", "_acme-challenge"]

        for fqdn in data:
            self.assertTrue(self.fqdn_with_leading_underscore.validate(fqdn), fqdn)

    def test_validate_underscore_negative(self):
        data = ["_acme-challenge._example.com"]

        for fqdn in data:
            self.assertFalse(self.fqdn_with_leading_underscore.validate(fqdn), fqdn)

    def test_validate_middle_underscore(self):
        data = ["abc_def.example.com"]

        for fqdn in data:
            self.assertTrue(self.fqdn_with_middle_underscore.validate(fqdn), fqdn)

    def test_validate_middle_underscore_negative(self):
        data = [
            "www_.example.com",
            "_www.example.com",
            "www.example._com",
            "www.example.com_",
            "www.example.co_m",
            "news.ae_ro",
        ]

        for fqdn in data:
            self.assertFalse(self.fqdn_with_middle_underscore.validate(fqdn), fqdn)

    def test_validate_negative(self):
        data = [
            "тест.ру",
            "-fe.fe",
            "fe-.fe",
            "_fe.fe",
            "f_e.fe",
            "fe_.fe",
            "a..bc",
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com",
            "-ec2_35$160%210-253.us-west-2-.compute.amazonaws.com",
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com",
            "a23456789-123456789-123456789-123456789-123456789-123456789-1234.b23.com",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.com",
            "mx.gmail.com.",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.com",
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn.validate(fqdn), fqdn)

    def test_min_levels(self):
        data = [
            "test-me.me",
            "fe.fe",
            "a.bc",
            "1.2.3.4.com",
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn_2level.validate(fqdn), fqdn)

    def test_min_levels_negative(self):
        data = [
            "first-level",
            "aa",
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn_2level.validate(fqdn), fqdn)


class FQDNTest(unittest.TestCase):
    def setUp(self):
        super(FQDNTest, self).setUp()
        self.fqdn = types_network.FQDN()
        self.fqdn_2level = types_network.FQDN(min_levels=2)

    def test_validate(self):
        data = [
            "first_level.",
            "test_me.me.",
            "fe.fe.",
            "aa.",
            "a.bc.",
            "1.2.3.4.com.",
            "xn--kxae4bafwg.xn--pxaix.gr.",
            "a23456789-123456789-123456789-123456789-123456789-123456789-123.b23.com.",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a234567.com.",
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk."
            "com.",
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn.validate(fqdn), fqdn)

    def test_validate_negative(self):
        data = [
            "тест.ру.",
            "-fe.fe.",
            "a..bc.",
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com.",
            "-ec2_35$160%210-253.us-west-2-.compute.amazonaws.com.",
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com.",
            "a23456789-123456789-123456789-123456789-123456789-123456789-1234.b23.com.",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.com.",
            "mx.gmail.com",
            "a23456789-a23456789-a234567890.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.a23456789.a23456789.a23456789.a23456789.a23456789."
            "a23456789.com.",
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn.validate(fqdn), fqdn)

    def test_min_levels(self):
        data = [
            "test_me.me.",
            "fe.fe.",
            "a.bc.",
            "1.2.3.4.com.",
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn_2level.validate(fqdn), fqdn)

    def test_min_levels_negative(self):
        data = [
            "first_level.",
            "aa.",
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn_2level.validate(fqdn), fqdn)


class IPRangeTest(unittest.TestCase):
    def setUp(self):
        self.ip_range = types_network.IPRange()

    def test_validate(self):
        correct_range = netaddr.IPRange("10.0.0.0", "10.0.1.0")
        self.assertTrue(self.ip_range.validate(correct_range))

        incorrect_range = "10.0.0.0-10.0.1.0"
        self.assertFalse(self.ip_range.validate(incorrect_range))

    def test_to_simple_type(self):
        foo_range = netaddr.IPRange("10.0.0.0", "10.0.1.0")
        self.assertEqual(self.ip_range.to_simple_type(foo_range), "10.0.0.0-10.0.1.0")

    def test_from_simple_type(self):
        foo_range = netaddr.IPRange("10.0.0.0", "10.0.1.0")
        self.assertEqual(self.ip_range.from_simple_type("10.0.0.0-10.0.1.0"), foo_range)

    def test_from_unicode(self):
        foo_range = netaddr.IPRange("10.0.0.0", "10.0.1.0")
        self.assertEqual(self.ip_range.from_unicode("10.0.0.0-10.0.1.0"), foo_range)



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import copy
import ctypes
import datetime
import decimal
import orjson
import re
import sys
import time
import uuid

import email_validator

INFINITY = float("inf")
INFINITI = INFINITY  # TODO(d.burmistrov): remove this hack
UUID_RE_TEMPLATE = (
    r"[0-9a-fA-F]{8}"
    r"-"
    r"[0-9a-fA-F]{4}"
    r"-"
    r"[0-9a-fA-F]{4}"
    r"-"
    r"[0-9a-fA-F]{4}"
    r"-"
    r"[0-9a-fA-F]{12}"
)

TIMEDELTA_INFINITY = (1 << (ctypes.sizeof(ctypes.c_int()) * 8) - 1) - 1

# Copy-paste from validators library because RA must support python 2.7
# and support cyrillic domain names. The validators library is located:
# https://github.com/kvesteri/validators/blob/master/validators/domain.py#L5
# the regexp has issue https://github.com/kvesteri/validators/issues/185
HOSTNAME_RE_TEMPLATE = (
    # First character of the domain
    "^(?:[a-zA-Z0-9]"
    # Sub domain + hostname
    "(?:[a-zA-Z0-9-_]{0,61}[A-Za-z0-9])?\.)"  # noqa
    # First 61 characters of the gTLD
    "+[A-Za-z0-9][A-Za-z0-9-_]{0,61}"
    # Last character of the gTLD
    "[A-Za-z]$"
)
KWARGS_OPENAPI_MAP = {
    "read_only": "readOnly",
    "default": "default",
    "example": "example",
}
MYSQL_DATETIME_FMT = "%Y-%m-%d %H:%M:%S.%f"
DEFAULT_DATE = datetime.datetime.strptime(
    "2006-01-02 15:04:05.000576", MYSQL_DATETIME_FMT
)
DEFAULT_DATE_Z = DEFAULT_DATE.replace(tzinfo=datetime.timezone.utc)
# RFC3339
# python's datetime doesn't support nanosecond precision
# + now without timezone, example "2006-01-02T15:04:05.999999999Z07:00"
OPENAPI_DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"


def build_prop_kwargs(kwargs, to_simple_type=None):
    result = {}
    for k, v in KWARGS_OPENAPI_MAP.items():
        if k in kwargs.keys():
            value = kwargs[k]() if callable(kwargs[k]) else kwargs[k]
            if isinstance(value, (UUID, uuid.UUID, bool)):
                # No default value for uuid and bool
                continue
            elif isinstance(value, datetime.datetime):
                value = DEFAULT_DATE.strftime(OPENAPI_DATETIME_FMT)
            elif hasattr(value, "__dict__"):
                value = {
                    name: prop.get_property_type().to_simple_type(value[name])
                    for name, prop in value.properties.properties.items()
                }
            elif to_simple_type is not None:
                value = to_simple_type(value)

            result[v] = value
    return result


class BaseType(metaclass=abc.ABCMeta):
    def __init__(self, openapi_type="object", openapi_format=None):
        super(BaseType, self).__init__()
        self._openapi_type = openapi_type
        self._openapi_format = openapi_format

    @abc.abstractmethod
    def validate(self, value):
        pass

    @abc.abstractmethod
    def to_simple_type(self, value):
        pass

    def dump_value(self, value):
        return self.to_simple_type(value)

    @abc.abstractmethod
    def from_simple_type(self, value):
        pass

    @abc.abstractmethod
    def from_unicode(self, value):
        pass

    def __repr__(self):
        return self.__class__.__name__

    @property
    def ra_type(self):
        return self.__class__

    @property
    def openapi_type(self):
        return self._openapi_type

    @property
    def openapi_format(self):
        return self._openapi_format

    def to_openapi_spec(self, prop_kwargs):
        spec = {"type": self._openapi_type}
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.to_simple_type)
        )
        return spec


class BasePythonType(BaseType):
    def __init__(self, python_type, **kwargs):
        super(BasePythonType, self).__init__(**kwargs)
        self._python_type = python_type

    def validate(self, value):
        return isinstance(value, self._python_type)

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return self._python_type(value)


class Boolean(BasePythonType):
    def __init__(self):
        super(Boolean, self).__init__(bool, openapi_type="boolean")

    def from_simple_type(self, value):
        return bool(value)

    def from_unicode(self, value):
        return value.lower() in ["yes", "true", "1"]


class String(BasePythonType):
    def __init__(self, min_length=0, max_length=sys.maxsize, **kwargs):
        openapi_type = kwargs.pop("openapi_type", "string")
        super(String, self).__init__(str, openapi_type=openapi_type, **kwargs)
        self.min_length = int(min_length)
        self.max_length = int(max_length)

    def validate(self, value):
        result = super(String, self).validate(value)
        return result and self.min_length <= len(str(value)) <= self.max_length

    def from_unicode(self, value):
        return str(value)

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "minLength": self.min_length,
            "maxLength": self.max_length,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Email(String):
    def __init__(
        self,
        min_length=5,
        max_length=254,
        check_deliverability=False,
        **kwargs,
    ):
        """
        Email type.

        :param min_length: Minimum length of email
        :param max_length: Maximum length of email
        :param check_deliverability: check email deliverability, default: False
        :param \\*\\*kwargs: Additional keyword arguments
        """
        openapi_type = kwargs.pop("openapi_type", "string")
        openapi_format = kwargs.pop("openapi_format", "email")
        super(Email, self).__init__(
            min_length=min_length,
            max_length=max_length,
            openapi_type=openapi_type,
            openapi_format=openapi_format,
            **kwargs,
        )
        self._check_deliverability = check_deliverability

    def validate(self, value):
        """
        Validates given value as an email address.

        :param value: Value to validate
        :return: True if value is valid, False otherwise
        """
        result = super(Email, self).validate(value)
        try:
            email_validator.validate_email(
                value,
                check_deliverability=self._check_deliverability,
            )
        except email_validator.EmailNotValidError:
            return False
        return result


class Integer(BasePythonType):
    def __init__(self, min_value=-INFINITY, max_value=INFINITY):
        super(Integer, self).__init__(int, openapi_type="integer")
        self.min_value = min_value if min_value == -INFINITY else int(min_value)
        self.max_value = max_value if max_value == INFINITY else int(max_value)

    def validate(self, value):
        result = super(Integer, self).validate(value)
        return result and self.min_value <= value <= self.max_value

    def from_unicode(self, value):
        return int(value)

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITI:
            return sys.maxsize
        else:
            return self.max_value

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITI:
            return -sys.maxsize
        else:
            return self.min_value

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Int8(Integer):
    def __init__(self):
        super(Int8, self).__init__(min_value=0, max_value=2**8 - 1)


class Float(BasePythonType):
    def __init__(self, min_value=-INFINITY, max_value=INFINITY):
        super(Float, self).__init__(
            float, openapi_type="number", openapi_format="float"
        )
        self.min_value = min_value if min_value == -INFINITY else float(min_value)
        self.max_value = max_value if max_value == INFINITY else float(max_value)

    def validate(self, value):
        result = super(Float, self).validate(value)
        return result and self.min_value <= value <= self.max_value

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITI:
            return sys.float_info.max
        else:
            return self.max_value

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITI:
            return -sys.float_info.max
        else:
            return self.min_value

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "format": self.openapi_format,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Decimal(BasePythonType):
    def __init__(
        self, min_value=-INFINITY, max_value=INFINITY, max_decimal_places=None
    ):
        super().__init__(
            python_type=decimal.Decimal,
            openapi_type="string",
            openapi_format="decimal",
        )
        self.min_value = decimal.Decimal(min_value)
        self.max_value = decimal.Decimal(max_value)
        self.max_decimal_places = max_decimal_places

    def validate(self, value):
        return (
            isinstance(value, decimal.Decimal)
            and self.min_value <= value <= self.max_value
            and (
                (self.max_decimal_places is None)
                or -value.as_tuple().exponent <= self.max_decimal_places
            )
        )

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return decimal.Decimal(str(value))

    def from_unicode(self, value):
        return self.from_simple_type(value)

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITY:
            return sys.maxsize
        else:
            return int(self.max_value)

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITY:
            return -sys.maxsize
        else:
            return int(self.min_value)

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.to_simple_type)
        )
        return spec


class UUID(BaseType):
    def __init__(self):
        super(UUID, self).__init__(openapi_type="string", openapi_format="uuid")

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        if isinstance(value, uuid.UUID):
            return value
        return uuid.UUID(value)

    def validate(self, value):
        return isinstance(value, uuid.UUID)

    def from_unicode(self, value):
        return uuid.UUID(value)


class ComplexPythonType(BasePythonType):
    _TYPE_ERROR_MSG = "Can't convert '%s' with type '%s' into %s"

    def _raise_on_invalid_type(self, value):
        if not isinstance(value, self._python_type):
            raise TypeError(
                self._TYPE_ERROR_MSG % (value, type(value), self._python_type)
            )

    def from_simple_type(self, value):
        self._raise_on_invalid_type(value)
        return value

    def from_unicode(self, value):
        result = None
        try:
            result = orjson.loads(value)
        except (TypeError, ValueError):
            pass
        self._raise_on_invalid_type(value)
        return self.from_simple_type(result)


class List(ComplexPythonType):
    def __init__(self):
        super(List, self).__init__(list, openapi_type="array", openapi_format="string")

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "format": self.openapi_format,
            "items": {"type": "string"},
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class TypedList(List):
    def __init__(self, nested_type):
        super(TypedList, self).__init__()
        if not isinstance(nested_type, BaseType):
            raise TypeError(
                "Nested type '%s' is not inherited from %s" % (nested_type, BaseType)
            )
        self._nested_type = nested_type

    def validate(self, value):
        return super(TypedList, self).validate(value) and all(
            self._nested_type.validate(item) for item in value
        )

    def to_simple_type(self, value):
        return [self._nested_type.to_simple_type(e) for e in value]

    def from_simple_type(self, value):
        return [self._nested_type.from_simple_type(e) for e in value]

    def from_unicode(self, value):
        if not isinstance(value, str):
            raise TypeError("Value must be str, not %s", type(value))

        value = self._nested_type.from_unicode(value)
        return [value]

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "items": self._nested_type.to_openapi_spec({}),
        }
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.to_simple_type)
        )
        return spec


class Dict(ComplexPythonType):
    def __init__(self):
        super(Dict, self).__init__(dict)

    def validate(self, value):
        return super(Dict, self).validate(value) and all(
            isinstance(k, str) for k in value
        )

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": "object",
            "additionalProperties": {
                "oneOf": [
                    {"type": "string"},
                    {"type": "integer"},
                    {"type": "boolean"},
                    {"type": "object"},
                    {"type": "array", "items": {}},
                ]
            },
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


def _validate_scheme(scheme):
    non_string_keys = []
    invalid_types = []

    for key, val in scheme.items():
        if not isinstance(key, str):
            non_string_keys.append(key)
        if not isinstance(val, BaseType):
            invalid_types.append(val)

    if non_string_keys:
        raise ValueError("Scheme keys %r are not strings" % non_string_keys)
    if invalid_types:
        raise ValueError(
            "Scheme values %r are not %s instances" % (invalid_types, BaseType)
        )


# TODO(d.burmistrov): we have to make this group of Dict Schemers:
#   - ExactSchemaDict - data must follow schema in every detail
#   - PartialSchemaDict - data must be within schema definition (some keys
#                         may be missing)
#   - ExtraSchemaDict - data must follow schema but may have extra keys
#   - not sure about this option: there may be extra keys (not defined
#     in schema) and some schema keys may be missing, but all defined keys
#     matching schema must be valid due to schema


class SoftSchemeDict(Dict):
    def __init__(self, scheme):
        super(SoftSchemeDict, self).__init__()
        _validate_scheme(scheme)
        self._scheme = scheme

    def validate(self, value):
        return (
            super(SoftSchemeDict, self).validate(value)
            and set(value.keys()).issubset(set(self._scheme.keys()))
            and all(self._scheme[k].validate(v) for k, v in value.items())
        )

    def to_simple_type(self, value):
        return {k: self._scheme[k].to_simple_type(v) for k, v in value.items()}

    def from_simple_type(self, value):
        value = super(SoftSchemeDict, self).from_simple_type(value)
        return {k: self._scheme[k].from_simple_type(v) for k, v in value.items()}

    def to_openapi_spec(self, prop_kwargs):
        spec = {"type": "object", "properties": {}}
        for k, v in self._scheme.items():
            spec["properties"][k] = v.to_openapi_spec(prop_kwargs)
        return spec


class SchemeDict(Dict):
    def __init__(self, scheme):
        super(SchemeDict, self).__init__()
        _validate_scheme(scheme)
        self._scheme = scheme

    def validate(self, value):
        return (
            super(SchemeDict, self).validate(value)
            and set(value.keys()) == set(self._scheme.keys())
            and all(scheme.validate(value[key]) for key, scheme in self._scheme.items())
        )

    def to_simple_type(self, value):
        return {
            key: scheme.to_simple_type(value[key])
            for key, scheme in self._scheme.items()
        }

    def from_simple_type(self, value):
        value = super(SchemeDict, self).from_simple_type(value)
        return {
            key: scheme.from_simple_type(value.get(key, None))
            for key, scheme in self._scheme.items()
        }

    def to_openapi_spec(self, prop_kwargs):
        spec = {"type": "object", "properties": {}}
        for k, v in self._scheme.items():
            spec["properties"][k] = v.to_openapi_spec(prop_kwargs)
        return spec


class TypedDict(Dict):
    def __init__(self, nested_type):
        super(TypedDict, self).__init__()
        if not isinstance(nested_type, BaseType):
            raise TypeError(
                "Nested type '%s' is not inherited from %s" % (nested_type, BaseType)
            )
        self._nested_type = nested_type

    def validate(self, value):
        return super(TypedDict, self).validate(value) and all(
            self._nested_type.validate(element) for element in value.values()
        )

    def to_simple_type(self, value):
        return {k: self._nested_type.to_simple_type(v) for k, v in value.items()}

    def from_simple_type(self, value):
        value = super(TypedDict, self).from_simple_type(value)
        return {k: self._nested_type.from_simple_type(v) for k, v in value.items()}

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self._openapi_type,
            "additionalProperties": self._nested_type.to_openapi_spec({}),
        }
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.to_simple_type)
        )
        return spec


class UTCDateTime(BasePythonType):
    """Deprecated utc datetime type. Use UTCDateTimeZ instead.

    UTCDateTime should be used only for compatibility,
    when you use naive datetime objects without datetimes.
    It's strongly recommended to use UTCDateTimeZ.
    """

    def __init__(self):
        super(UTCDateTime, self).__init__(
            python_type=datetime.datetime,
            openapi_type="string",
            openapi_format="date-time",
        )

    def validate(self, value):
        return isinstance(value, datetime.datetime) and (
            value.tzinfo == datetime.timezone.utc or value.tzinfo is None
        )

    def to_simple_type(self, value):
        return value.strftime(MYSQL_DATETIME_FMT)

    def dump_value(self, value):
        # Converting value in api response
        return value.strftime(OPENAPI_DATETIME_FMT)

    def from_simple_type(self, value):
        if isinstance(value, datetime.datetime):
            return value
        try:
            return datetime.datetime.strptime(value, MYSQL_DATETIME_FMT)
        except ValueError:
            # Used in cases, than we manually convert openapi string in
            # http response to ra type in model
            return datetime.datetime.strptime(value, OPENAPI_DATETIME_FMT)

    def from_unicode(self, value):
        return self.from_simple_type(value)

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self._openapi_type,
            # https://github.com/ogen-go/ogen/blob/main/_testdata/positive/time_extension.yml#L29
            "x-ogen-time-format": self.dump_value(DEFAULT_DATE),
        }
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.dump_value)
        )
        return spec


class UTCDateTimeZ(UTCDateTime):
    """Appropriate datetime UTC type with guarantees for tzinfo existence."""

    def validate(self, value):
        return isinstance(value, datetime.datetime) and (
            value.tzinfo == datetime.timezone.utc
        )

    def from_simple_type(self, value):
        result = super(UTCDateTimeZ, self).from_simple_type(value)
        if result.tzinfo is not None:
            return result.astimezone(datetime.timezone.utc)
        # If datetime is naive, it's assumed that timezone is UTC, add it
        return result.replace(tzinfo=datetime.timezone.utc)


class TimeDelta(BasePythonType):
    """Appropriate timedelta type."""

    def __init__(self, min_value=-TIMEDELTA_INFINITY, max_value=TIMEDELTA_INFINITY):
        self._min_value = min_value
        self._max_value = max_value
        super().__init__(
            python_type=datetime.timedelta,
            openapi_type="number",
            openapi_format="float",
        )

    @property
    def min_value(self):
        return self._min_value

    @property
    def max_value(self):
        return self._max_value

    def validate(self, value):
        result = super().validate(value)
        return result and self.min_value <= self.to_simple_type(value) <= self.max_value

    def to_simple_type(self, value):
        return value.total_seconds()

    def from_simple_type(self, value):
        if isinstance(value, datetime.timedelta):
            return value
        return datetime.timedelta(seconds=value)

    def from_unicode(self, value):
        return self.from_simple_type(value)

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITI:
            return sys.float_info.max
        else:
            return self.max_value

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITI:
            return -sys.float_info.max
        else:
            return self.min_value

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "format": self.openapi_format,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        spec.update(
            build_prop_kwargs(kwargs=prop_kwargs, to_simple_type=self.to_simple_type)
        )
        return spec


class DateTime(BasePythonType):
    def __init__(self, min_value=None, max_value=None):
        super(DateTime, self).__init__(python_type=datetime.datetime)

    def to_simple_type(self, value):
        return int(time.mktime(value.utctimetuple()))

    def from_simple_type(self, value):
        return datetime.datetime.utcfromtimestamp(value)

    def from_unicode(self, value):
        return self.from_simple_type(value)


class Enum(BaseType):
    def __init__(self, enum_values):
        super(Enum, self).__init__(openapi_type="string")
        self._enums_values = copy.deepcopy(enum_values)

    @property
    def values(self):
        return self._enums_values

    def validate(self, value):
        return value in self._enums_values

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        for enum_value in self._enums_values:
            if value == str(enum_value):
                return enum_value
        raise TypeError(
            "Can't convert '%s' to enum type."
            " Allowed values are %s" % (value, self._enums_values)
        )

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "enum": sorted(list(self.values)),
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class BaseRegExpType(BaseType):
    """BaseCompiledRegExpTypeFromAttr is preferred to be used"""

    def __init__(self, pattern, openapi_type="string", **kwargs):
        super(BaseRegExpType, self).__init__(openapi_type, **kwargs)
        self._pattern = re.compile(pattern)

    def validate(self, value):
        try:
            return self._pattern.match(value) is not None
        except TypeError:
            return False

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return value

    @property
    def pattern_openapi(self):
        return self._pattern.pattern

    def to_openapi_spec(self, prop_kwargs):
        spec = {"type": self.openapi_type, "pattern": self.pattern_openapi}
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class BaseCompiledRegExpType(BaseRegExpType):
    def __init__(self, pattern, **kwargs):
        if "openapi_type" not in kwargs:
            kwargs["openapi_type"] = "string"
        super(BaseRegExpType, self).__init__(**kwargs)
        self._pattern = pattern


class BaseCompiledRegExpTypeFromAttr(BaseCompiledRegExpType):
    def __init__(self, **kwargs):
        super(BaseCompiledRegExpTypeFromAttr, self).__init__(
            pattern=self.pattern, **kwargs
        )


class Uri(BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^(/[A-Za-z0-9\-_]*)*/%s$" % UUID_RE_TEMPLATE)

    def __init__(self):
        super(Uri, self).__init__(
            openapi_type="string",
            openapi_format="uri",
        )


class Mac(BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^([0-9a-fA-F]{2,2}:){5,5}[0-9a-fA-F]{2,2}$")

    def __init__(self):
        super(Mac, self).__init__(
            openapi_type="string",
            openapi_format="mac",
        )


class Hostname(BaseCompiledRegExpTypeFromAttr):
    """DEPRECATED! Use types from types_network module"""

    pattern = re.compile(HOSTNAME_RE_TEMPLATE)

    def __init__(self):
        super(Hostname, self).__init__(
            openapi_type="hostname",
        )


class Url(BaseCompiledRegExpTypeFromAttr):
    """

    django url validation regex
    (https://github.com/django/django/blob/stable/1.3.x/django/core/validators.py#L45):
    """

    pattern = re.compile(
        r"^(?:http|ftp)s?://"
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # noqa
        r"localhost|"
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
        r"(?::\d+)?"
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )


class AllowNone(BaseType):
    def __init__(self, nested_type):
        super(AllowNone, self).__init__(openapi_type="string")
        self._nested_type = nested_type

    @property
    def nested_type(self):
        return self._nested_type

    def validate(self, value):
        return value is None or self._nested_type.validate(value)

    def to_simple_type(self, value):
        return None if value is None else self._nested_type.to_simple_type(value)

    def from_simple_type(self, value):
        return None if value is None else self._nested_type.from_simple_type(value)

    def from_unicode(self, value):
        if value == "null":
            return None
        else:
            return self._nested_type.from_unicode(value)

    def to_openapi_spec(self, prop_kwargs):
        if "default" in prop_kwargs and prop_kwargs["default"] is None:
            del prop_kwargs["default"]
        spec = self._nested_type.to_openapi_spec(prop_kwargs)
        spec["nullable"] = True
        return spec


class AnySimpleType(BasePythonType):
    """Accepts any simple type.

    Example:
    >>> AnySimpleType().validate(1)
    True
    >>> AnySimpleType().validate("foo")
    True
    >>> AnySimpleType().validate([1, 2, 3])
    True
    >>> AnySimpleType().validate({"foo": "bar"})
    True
    >>> AnySimpleType().validate(True)
    True
    """

    def __init__(self):
        self._simple_types = (int, float, list, dict, bool, str)
        super().__init__(python_type=self._simple_types)

    def from_unicode(self, value: str | bytes):
        try:
            # Handles JSON-encoded numbers, booleans, lists,
            # dicts, and strings.
            return orjson.loads(value)
        except orjson.JSONDecodeError:
            raise TypeError(f"Incorrect value {value} for type {type(self)}")



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# HTTP methods
GET = "GET"
PUT = "PUT"
POST = "POST"

# RA methods
FILTER = "FILTER"
CREATE = "CREATE"
UPDATE = "UPDATE"
DELETE = "DELETE"

# RA action methods
ACTION_GET = "ACTION_GET"
ACTION_POST = "ACTION_POST"
ACTION_PUT = "ACTION_PUT"

# all RA methods
ALL = "ALL"

ALL_RA_METHODS = [
    GET,  # Controller.get
    FILTER,  # Controller.filter
    CREATE,  # Controller.create
    UPDATE,  # Controller.update
    DELETE,  # Controller.delete
    ACTION_GET,  # Action.get
    ACTION_POST,  # Action.post
    ACTION_PUT,  # Action.put
    ALL,  # All RA methods + actions
]


HTTP_TO_RA_COLLECTION_METHODS = {GET: FILTER, POST: CREATE}
HTTP_TO_RA_RESOURCE_METHODS = {GET: GET, PUT: UPDATE, DELETE: DELETE}

CONTENT_TYPE_APPLICATION_JSON = DEFAULT_CONTENT_TYPE = "application/json"
CONTENT_TYPE_MULTIPART = "multipart/form-data"
CONTENT_TYPE_OCTET_STREAM = "application/octet-stream"



# Copyright (c) 2014 Eugene Frolov <efrolov@mirantis.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import sys
import traceback

from restalchemy.api import middlewares

LOG = logging.getLogger(__name__)


class LoggingMiddleware(middlewares.Middleware):
    """API logging middleware."""

    SENSITIVE_HEADERS = {
        "AUTHORIZATION",
    }

    def __init__(self, application, logger_name=__name__):
        super(LoggingMiddleware, self).__init__(application)
        self.logger = logging.getLogger(logger_name)

    def process_request(self, req):
        req_chunk = self._request_chunk(req)
        sanitized_headers = self._sanitize_headers(req.headers)
        self.logger.debug(
            "API > %s headers=%s body=%r",
            req_chunk,
            self._headers_chunk(sanitized_headers),
            req.body,
        )
        try:
            res = req.get_response(self.application)
            # XXX(Eugeny Flolov):
            # :py:method:`middlewares.ContextMiddleware#process_response`
            # unreachable if
            # :py:method:`middlewares.ContextMiddleware#process_request`
            # returns response.
            self.logger.debug(
                "API < %s %s headers=%s body=%r",
                res.status_code,
                self._request_chunk(req),
                self._headers_chunk(res.headers),
                res.body,
            )
            return res
        except Exception:
            e_type, e_value, e_tb = sys.exc_info()
            e_file, e_lineno, e_fn, e_line = traceback.extract_tb(e_tb)[-1]
            self.logger.error(
                "API Error %s %s %s %s:%s:%s> %s",
                req_chunk,
                e_type,
                e_value,
                e_file,
                e_lineno,
                e_fn,
                e_line,
            )
            raise

    @staticmethod
    def _request_chunk(req):
        return "%s %s" % (req.method, req.url)

    @staticmethod
    def _headers_chunk(headers):
        return ["%s: %s" % (h, headers[h]) for h in headers]

    def _sanitize_headers(self, headers):
        def _sanitized(header, header_value):
            if str(header).upper() in self.SENSITIVE_HEADERS:
                return "***"
            return header_value

        return {k: _sanitized(k, v) for k, v in headers.items()}



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm


# CREATE TABLE `foos` (
#      `uuid` CHAR(36) CHARACTER SET utf8 COLLATE utf8_unicode_ci NOT NULL ,
#      `foo_field1` INT NOT NULL ,
#      `foo_field2` VARCHAR(255) NOT NULL ,
# PRIMARY KEY (`uuid`(36)) USING HASH)
# ENGINE = InnoDB;
class FooModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "foos"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="field2_value")


engines.engine_factory.configure_factory(
    db_url="mysql://root:21070809d@127.0.0.1/test",
)


model1 = FooModel(foo_field1=10, foo_field2="Value1")
model2 = FooModel(foo_field1=20, foo_field2="Value2")
model3 = FooModel(foo_field1=30, foo_field2="TypeA")
model1.save()
model2.save()
model3.save()

print(FooModel.objects.count())
print(FooModel.objects.count(filters={"foo_field1": filters.GT(10)}))
print(FooModel.objects.count(filters={"foo_field2": "TypeA"}))

for model in FooModel.objects.get_all():
    model.delete()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm


# CREATE TABLE foos (
#      uuid CHAR(36) PRIMARY KEY,
#      foo_field1 INT NOT NULL ,
#      foo_field2 VARCHAR(255) NOT NULL
# );
class FooModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "foos"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


# CREATE TABLE bars (
#     uuid CHAR(36) NOT NULL PRIMARY KEY,
#     bar_field1 VARCHAR(10) NOT NULL,
#     foo CHAR(36) NOT NULL,
#     FOREIGN KEY (foo) REFERENCES foos(uuid) ON DELETE RESTRICT ON UPDATE
#       RESTRICT
# );
#
# CREATE INDEX rlfoo ON bars(foo);
class BarModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "bars"
    bar_field1 = properties.property(types.String(min_length=1, max_length=10))
    foo = relationships.relationship(FooModel)


engines.engine_factory.configure_factory(
    db_url="postgresql://posgres:password@127.0.0.1:5432/ra_tests"
)

# Create new foo object and store it
foo1 = FooModel(foo_field1=10)
foo1.save()

bar1 = BarModel(bar_field1="test", foo=foo1)
bar1.save()

print(list(BarModel.objects.get_all()))
print(BarModel.objects.get_one(filters={"uuid": bar1.get_id()}))
print(list(BarModel.objects.get_all(filters={"foo": foo1})))
print(bar1.as_plain_dict())

bar1.delete()

foo2 = FooModel(foo_field1=11, foo_field2="some text")
foo2.save()

foos = list(FooModel.objects.get_all())
print(foos)

print(FooModel.objects.get_one(filters={"foo_field1": filters.EQ(10)}))

# Modify foo_field2 and update it in storage
foo2.foo_field2 = "xxx2 asdad asdasd"
foo2.save()

# Delete foo object from storage
for foo in foos:
    foo.delete()

print("foo_field1 is greater than 5")
for num in range(10):
    foo = FooModel(foo_field1=num)
    foo.save()

print(list(FooModel.objects.get_all(filters={"foo_field1": filters.GT(5)})))

print("foo_field1 in equal 5 or 6")
print(FooModel.objects.get_all(filters={"foo_field1": filters.In([5, 6])}))
print(FooModel.objects.get_all(filters={"foo_field1": filters.NotIn([1, 2])}))

for model in FooModel.objects.get_all():
    model.delete()

# Complex filters
# WHERE ((`foo_field1` = 1 AND `foo_field2` = 2) OR (`foo_field2` = 3))
# filter_list = filters.OR(
#     filters.AND(
#         {
#             "foo_field1": filters.EQ(1),
#             "foo_field2": filters.EQ("2"),
#         }
#     ),
#     filters.AND({"foo_field2": filters.EQ("3")}),
# )

# FooModel(foo_field1=1, foo_field2="2").save()

# print(FooModel.objects.get_one(filters=filter_list))



# Copyright 2022 George Melikov
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
import webob

from restalchemy.api import constants
from restalchemy.api import contexts
from restalchemy.api import field_permissions as fp
from restalchemy.api import resources
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types


class FakeModel(models.CustomPropertiesMixin, models.ModelWithUUID):
    _private_field = properties.property(types.Integer())
    standard_field1 = properties.property(types.Integer(), required=True)
    standard_field2 = properties.property(types.Integer())
    standard_field3 = properties.property(types.Integer())
    standard_field4 = properties.property(types.Integer(), required=True)
    standard_field5 = properties.property(types.Integer())


class FakeMemberContext(object):
    roles = ["_member_"]


class FakeAdminContext(object):
    roles = ["admin"]


class FakeEmptyContext(object):
    roles = []


class ResourceByRAModelFieldsPermissions(unittest.TestCase):
    def setUp(self):
        super(ResourceByRAModelFieldsPermissions, self).setUp()

        self.resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=fp.FieldsPermissionsByRole(
                default=fp.UniversalPermissions(permission=fp.Permissions.HIDDEN),
                admin=fp.FieldsPermissions(
                    {
                        "standard_field1": {constants.ALL: fp.Permissions.HIDDEN},
                        "standard_field2": {constants.CREATE: (fp.Permissions.HIDDEN)},
                        "standard_field3": {constants.GET: fp.Permissions.HIDDEN},
                        "standard_field5": {
                            constants.FILTER: (fp.Permissions.HIDDEN),
                            constants.DELETE: (fp.Permissions.HIDDEN),
                        },
                    }
                ),
                _member_=fp.FieldsPermissions(
                    {
                        "standard_field1": {
                            constants.CREATE: (fp.Permissions.HIDDEN),
                            constants.ALL: fp.Permissions.RO,
                        },
                        "standard_field2": {
                            constants.GET: fp.Permissions.RO,
                            constants.FILTER: fp.Permissions.RO,
                            constants.ALL: fp.Permissions.HIDDEN,
                        },
                        "standard_field4": {
                            constants.GET: fp.Permissions.RO,
                            constants.UPDATE: fp.Permissions.RO,
                            constants.ALL: fp.Permissions.HIDDEN,
                        },
                        "standard_field5": {
                            constants.UPDATE: (fp.Permissions.HIDDEN),
                            constants.DELETE: (fp.Permissions.HIDDEN),
                        },
                    }
                ),
            ),
        )

        self._request = webob.Request.blank("/some-uri")
        self._request.api_context = contexts.RequestContext(self._request)

    def tearDown(self):
        super(ResourceByRAModelFieldsPermissions, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._request

    def _test_fields_is_shown(self, expected_fields):
        result = [
            name
            for name, prop in self.resource.get_fields_by_request(self._request)
            if prop.is_public()
            and not self.resource._fields_permissions.is_hidden(name, self._request)
        ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_fields_is_hidden(self, expected_fields):
        result = [
            name
            for name, prop in self.resource.get_fields_by_request(self._request)
            if prop.is_public()
            and self.resource._fields_permissions.is_hidden(name, self._request)
        ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_fields_is_readonly(self, expected_fields):
        result = [
            name
            for name, prop in self.resource.get_fields_by_request(self._request)
            if prop.is_public()
            and self.resource._fields_permissions.is_readonly(name, self._request)
        ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_default(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=fp.FieldsPermissionsByRole(
                default=fp.UniversalPermissions(permission=fp.Permissions.HIDDEN),
                admin=fp.FieldsPermissions(
                    {
                        "standard_field1": {constants.ALL: fp.Permissions.HIDDEN},
                    },
                ),
            ),
        )

        self.assertFalse(
            resource._fields_permissions.is_readonly("standard_field2", self._request)
        )

    def _test_default_custom_value(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=fp.FieldsPermissionsByRole(
                default=fp.UniversalPermissions(permission=fp.Permissions.HIDDEN),
                admin=fp.FieldsPermissions(
                    {
                        "standard_field1": {constants.ALL: fp.Permissions.HIDDEN},
                    },
                    default=fp.Permissions.RO,
                ),
            ),
        )

        self.assertTrue(
            resource._fields_permissions.is_readonly("standard_field2", self._request)
        )


class ResourceByRAModelFieldsPermissionsRoleAdmin(ResourceByRAModelFieldsPermissions):
    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsRoleAdmin, self).setUp()
        admin_context = FakeAdminContext()
        self._request.context = admin_context

    def test_fields_permission_role_admin_method_get(self):
        self._request.api_context.set_active_method(constants.GET)

        expected_shown_fields = [
            "standard_field2",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1", "standard_field3"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1", "standard_field5"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)

        expected_shown_fields = [
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = [
            "standard_field1",
            "standard_field2",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1", "standard_field5"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)

        expected_shown_fields = [
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field1"]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)


class ResourceByRAModelFieldsPermissionsNoRole(ResourceByRAModelFieldsPermissions):
    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsNoRole, self).setUp()
        empty_context = FakeEmptyContext()
        self._request.context = empty_context

        self.expected_shown_fields = []
        self.expected_hidden_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]

    def test_fields_permission_no_role_method_get(self):
        self._request.api_context.set_active_method(constants.GET)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)


class ResourceByRAModelFieldsPermissionsRoleMember(ResourceByRAModelFieldsPermissions):
    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsRoleMember, self).setUp()
        member_context = FakeMemberContext()
        self._request.context = member_context

    def test_fields_permission_role_member_method_get(self):
        self._request.api_context.set_active_method(constants.GET)
        expected_shown_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field3",
            "standard_field4",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = []
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)
        expected_shown_fields = [
            "standard_field1",
            "standard_field3",
            "standard_field4",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field2", "standard_field5"]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
            "standard_field5",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)
        expected_shown_fields = ["standard_field3", "standard_field5", "uuid"]
        expected_hidden_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_member_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)
        expected_shown_fields = ["standard_field1", "standard_field3", "uuid"]
        expected_hidden_fields = [
            "standard_field2",
            "standard_field4",
            "standard_field5",
        ]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
            "standard_field5",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)
        expected_shown_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field3",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field4"]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)
        expected_shown_fields = [
            "standard_field1",
            "standard_field3",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field2", "standard_field4"]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)
        expected_shown_fields = [
            "standard_field1",
            "standard_field3",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field2", "standard_field4"]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)
        expected_shown_fields = [
            "standard_field1",
            "standard_field3",
            "standard_field5",
            "uuid",
        ]
        expected_hidden_fields = ["standard_field2", "standard_field4"]
        expected_ro_fields = [
            "standard_field1",
            "standard_field2",
            "standard_field4",
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "502944fe-e739-4c1a-a413-24fa787ae14d"

    def upgrade(self, session):
        sql_expr_list = [
            """
                CREATE TABLE test_count (
                    uuid CHAR(36) NOT NULL,
                    foo_field1 INT NOT NULL,
                    foo_field2 VARCHAR(255) NOT NULL,
                PRIMARY KEY (uuid));
            """,
            """
                INSERT INTO test_count (
                     uuid, foo_field1, foo_field2
                ) VALUES (
                    '00000000-0000-0000-0000-000000000000', 1, 'value1'
                ), (
                    '00000000-0000-0000-0000-000000000001', 2, 'value2'
                ), (
                    '00000000-0000-0000-0000-000000000002', 3, 'value3'
                ), (
                    '00000000-0000-0000-0000-000000000003', 4, 'other1'
                )
            """,
        ]

        [session.execute(expr) for expr in sql_expr_list]

    def downgrade(self, session):
        self._delete_table_if_exists(session, "test_count")


migration_step = MigrationStep()



# Copyright 2023 v.burygin
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import http.client as http_client

from restalchemy.api.middlewares import errors
from restalchemy.common import exceptions as comm_exc
from restalchemy.storage import exceptions as ra_exceptions
from restalchemy.tests.unit import base

from webob import request


class FakeResponse(object):
    def __init__(self, status, json, **kwargs):
        self.status = status
        self.json = json


class FakeConflictRecords(ra_exceptions.ConflictRecords):
    message = "Duplicate parameters"


class FakeRecordNotFound(ra_exceptions.RecordNotFound):
    message = "Can't found record in storage"


class ErrorsHandlerMiddlewareTestCase(base.BaseTestCase):
    middlew = errors.ErrorsHandlerMiddleware("application")

    def test_response_restalchemy_exc(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = comm_exc.RestAlchemyException
        request_mock.ResponseClass = FakeResponse
        response = self.middlew.process_request(request_mock)

        assert response.status == comm_exc.RestAlchemyException.code

    def test_response_validation_error(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = comm_exc.ValidationErrorException
        request_mock.ResponseClass = FakeResponse
        response = self.middlew.process_request(request_mock)

        assert response.status == http_client.BAD_REQUEST

    def test_response_conflict_record(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = FakeConflictRecords
        request_mock.ResponseClass = FakeResponse
        response = self.middlew.process_request(request_mock)

        assert response.status == http_client.CONFLICT

    def test_response_not_found(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = FakeRecordNotFound
        request_mock.ResponseClass = FakeResponse
        response = self.middlew.process_request(request_mock)

        assert response.status == http_client.NOT_FOUND

    def test_absent_message_attr_for_base_exceptions(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = TypeError
        request_mock.ResponseClass = FakeResponse
        response = self.middlew.process_request(request_mock)

        assert response.status == http_client.INTERNAL_SERVER_ERROR



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import inspect

from restalchemy.api import constants
from restalchemy.api import field_permissions
from restalchemy.common import exceptions as exc
from restalchemy.dm import properties as ra_properties
from restalchemy.dm import relationships as ra_relationsips


class ResourceMap(object):
    resource_map = {}
    model_type_to_resource = {}

    @classmethod
    def get_location(cls, model):
        resource = cls.get_resource_by_model(model)
        if resource not in cls.resource_map:
            raise exc.UnknownResourceLocation(resource=resource)
        return cls.resource_map[resource].get_uri(model)

    @classmethod
    def get_locator(cls, uri):
        for resource, locator in cls.resource_map.items():
            if locator.is_your_uri(uri):
                return locator
        raise exc.LocatorNotFound(uri=uri)

    @classmethod
    def get_resource(cls, request, uri):
        """Get any resource from service using request and custom uri

        This method allows to get any resource, if there is a controller with
        a get method for it. In this case, the request will be the same as the
        call through the API, with the exception of the passage of
        middlewares.

        :param request: The user request
        :param uri: The custom URI for desired resource

        :return: The resource which controller returns
        """
        resource_locator = cls.get_locator(uri)
        uri_stack = uri.split("/")

        # has parent resource?
        pstack = resource_locator.path_stack
        parent_resource = None

        # NOTE(efrolov): Get all resources except the last
        for num in range(len(pstack[:-1])):
            if not isinstance(pstack[num], str):
                # NOTE(efrolov): pstack is shorter than uri_stack by 1
                #                element. And I have to grab the ID, so +2.
                parent_uri = "/".join(uri_stack[0 : num + 2])
                parent_locator = cls.get_locator(parent_uri)
                parent_resource = parent_locator.get_resource(
                    request, parent_uri, parent_resource=parent_resource
                )

        return resource_locator.get_resource(request, uri, parent_resource)

    @classmethod
    def set_resource_map(cls, resource_map):
        cls.resource_map = resource_map

    @classmethod
    def add_model_to_resource_mapping(cls, model_class, resource):
        if model_class in cls.model_type_to_resource:
            raise ValueError(
                "model (%s) for resource (%s) already added. %s"
                % (model_class, resource, cls.model_type_to_resource)
            )
        cls.model_type_to_resource[model_class] = resource

    @classmethod
    def get_resource_by_model(cls, model):
        model_type = model.get_model_type()
        try:
            return cls.model_type_to_resource[model_type]
        except KeyError:
            raise exc.CanNotFindResourceByModel(model=model)


class AbstractResourceProperty(metaclass=abc.ABCMeta):
    def __init__(self, resource, model_property_name, public=True):
        super(AbstractResourceProperty, self).__init__()
        self._resource = resource
        self._model_property_name = model_property_name
        self._hidden = False
        self._public = public

    def is_public(self):
        return self._public

    def get_type(
        self,
    ):
        return self._resource.get_property_type(
            property_name=self._model_property_name,
        )

    def is_id_property(self):
        return self._resource.is_id_property(
            property_name=self._model_property_name,
        )

    @property
    def api_name(self):
        return self._resource.get_resource_field_name(self._model_property_name)

    @property
    def name(self):
        return self._model_property_name

    @abc.abstractmethod
    def parse_value(self, req, value):
        raise NotImplementedError()

    @abc.abstractmethod
    def parse_value_from_unicode(self, req, value):
        raise NotImplementedError()

    @abc.abstractmethod
    def dump_value(self, value):
        return NotImplementedError()


class ResourceProperty(AbstractResourceProperty):
    pass


class ResourceRAProperty(ResourceProperty):
    def __init__(self, resource, prop_type, model_property_name, public=True):
        super(ResourceRAProperty, self).__init__(
            resource=resource,
            model_property_name=model_property_name,
            public=public,
        )
        self._prop_type = prop_type() if inspect.isclass(prop_type) else prop_type

    def parse_value(self, req, value):
        return self._prop_type.from_simple_type(value)

    def parse_value_from_unicode(self, req, value):
        return self._prop_type.from_unicode(value)

    def dump_value(self, value):
        return self._prop_type.dump_value(value)


class ResourceRelationship(AbstractResourceProperty):
    def parse_value(self, req, value):
        return ResourceMap.get_resource(req, value)

    def parse_value_from_unicode(self, req, value):
        return self.parse_value(req, value)

    def dump_value(self, value):
        return ResourceMap.get_location(value)

    def is_id_property(self):
        return False


class BaseHiddenFieldsMap(object):
    def __init__(self, hidden_fields=None):
        super(BaseHiddenFieldsMap, self).__init__()
        self._hidden_fields = set(hidden_fields or [])

    @property
    def hidden_fields(self):
        return self._hidden_fields

    def is_hidden_field(self, model_field_name, req):
        return model_field_name in self

    def is_hidden_field_by_method(self, model_field_name, method):
        return model_field_name in self

    def __contains__(self, item):
        # NOTE(efrolov): backward compatibility
        return item in self._hidden_fields


class HiddenFieldsCompatibleClass(BaseHiddenFieldsMap):
    pass


class HiddenFieldMap(BaseHiddenFieldsMap):
    def __init__(self, **kwargs):
        """Hidden fields mapper for resource

        This class describes a list of fields that should be hidden for
        various RestAlchemy API methods. The methods supported by RA are
        declared in the `restalchemy.api.constants` module. To hide the
        `my_hidden_field` field for the FILTER method, the code will look
        like this:

        ```
        HiddenFieldMap(filter=['my_hidden_field'])
        ```

        For all other RA API methods, `my_hidden_field` field will not be
        hidden.

        :param filter: HTTP method GET for :func:`Controller.filter` method
        :param get: HTTP method GET for :func:`Controller.get` method
        :param create: HTTP method POST for :func:`Controller.create` method
        :param update: HTTP method PUT for :func:`Controller.update` method
        :param delete: HTTP method PUT for :func:`Controller.delete` method
        :param action_get: HTTP method GET for :func:`Action.get` method
        :param action_post: HTTP method POST for :func:`Action.post` method
        :param action_put: HTTP method PUT for :func:`Action.put` method
        """
        params = {}
        all_values = []
        for method in constants.ALL_RA_METHODS:
            value_arg = kwargs.pop(method.lower(), [])
            params[method] = value_arg
            all_values += value_arg
        if kwargs:
            raise TypeError("Got an unexpected keyword arguments %r" % kwargs)
        super(HiddenFieldMap, self).__init__(hidden_fields=all_values)
        self._method_map = {m: set(v) for m, v in params.items()}

    def is_hidden_field(self, model_field_name, req):
        """Checks that a field is in the list of hidden list

        :param model_field_name: The field name
        :param req: The webob request
        :return: True or False
        """
        try:
            method = req.api_context.get_active_method()
            return model_field_name in self._method_map[method]
        except KeyError:
            raise NotImplementedError("Unsupported RA method `%s`" % req)

    def is_hidden_field_by_method(self, model_field_name, method):
        try:
            return model_field_name in self._method_map[method]
        except KeyError:
            raise NotImplementedError("Unsupported RA method `%s`" % method)


class RoleBasedHiddenFieldContainer(BaseHiddenFieldsMap):
    def __init__(self, default, **kwargs):
        """The role based hidden field container

        The container of hidden fields. The class calculates the hidden fields
        using the roles from the oslo context.

        This class describes a list of fields that should be hidden for
        various oslo context roles. An object of the HiddenFieldMap class is
        used to specify lists of hidden fields for a role. The example below
        describes a different set of hidden fields for a user with the admin
        role and for users without the admin role:

        ```
        RoleBasedHiddenFieldContainer(
            default=HiddenFieldMap(get=['only_for_admin', 'hidden_field']),
            admin=HiddenFieldMap(get=['hidden_field']),
        )
        ```

        The example shows that the resource fields `only_for_admin` and
        `hidden_field` will be hidden by default for all roles. For the admin
        role, only `hidden_field` field is hidden.

        :param default: The instance of :class:`HiddenFieldMap` class. Hidden
                        fields for any role that has no other rules defined.
        :type default: HiddenFieldMap
        :param **kwargs: An optional parameter. The parameter name is the name
                         of the role name and parameter value is an instance
                         of :class:`HiddenFieldMap` class.
        :type default: HiddenFieldMap
        """
        self._default_hidden_fields = default
        self._hidden_fields_by_role = kwargs
        super(RoleBasedHiddenFieldContainer, self).__init__(
            hidden_fields=default.hidden_fields,
        )

    @staticmethod
    def _get_roles(req):
        """Returns the roles

        Returns the roles from the oslo context or an empty list if the
        context does not exist in the request from the user. Oslo context may
        be missing in the request if keystone middleware is not included to
        wsgi pipeline.

        :param req: The webob request that can contain oslo context
        :return: The list of roles from context or empty list if context is
                 missing.

        """
        roles = []

        if hasattr(req, "context") and hasattr(req.context, "roles"):
            roles = req.context.roles

        return roles

    def is_hidden_field(self, model_field_name, req):
        """Checks that a field is in the list of hidden list

        The field is considered hidden if the field is included to all hidden
        fields lists    for the specified roles.

        :param model_field_name: The field name
        :param req: The webob request that can contain oslo context
        :return: True or False
        """
        context_roles = self._get_roles(req)

        for rname, h_fields in self._hidden_fields_by_role.items():
            if rname in context_roles and not h_fields.is_hidden_field(
                model_field_name,
                req,
            ):
                return False
        return self._default_hidden_fields.is_hidden_field(model_field_name, req)

    def is_hidden_field_by_method(self, model_field_name, method):
        return True


class AbstractResource(metaclass=abc.ABCMeta):
    def __init__(
        self,
        model_class,
        name_map=None,
        hidden_fields=None,
        convert_underscore=True,
        process_filters=False,
        model_subclasses=None,
        fields_permissions=None,
    ):
        """Resource constructor

        :param model_class: The model class that is the source of the fields
                            for the resource
        :param name_map: The dictionary whose key is the name of the field in
                         the model and the value is the name of the field in
                         the resource. All model fields that match the names
                         from the passed keys in dictionary will be renamed to
                         values in passed dictionary.
        :param hidden_fields: The list of field names or instance of
                              :class:`HiddenFieldMap` class to hide from the
                              API user. The user will also not be able to set
                              these fields using API. All fields starting with
                              _ are already hidden from the user.
        :param convert_underscore: The boolean value. Should a resource
                                   convert _ to -
        :param process_filters: The boolean value. If the value is True then RA
                                will try to automatically parse the filters and
                                convert the filter values to the field type of
                                the model (resource).
        :param model_subclasses: The list of subclasses that can be represented
                                 by this resource, most often these are the
                                 children of the model specified in the
                                 model_class argument.
        :param fields_permissions: The dict of field and permissions, instance
                                 of `field_permissions.BasePermissions`.
                                 Use for setting field hidden or readonly by
                                 role from request context. if
                                 fields_permissionsis wouldn't set,
                                 it would be the object of UniversalPermissions
                                 with READWRITE permissions to all fields
        """
        super(AbstractResource, self).__init__()
        self._model_class = model_class
        self._name_map = name_map or {}
        self._inv_name_map = {v: k for k, v in self._name_map.items()}
        # NOTE(efrolov): to support the old resource interface
        if not isinstance(hidden_fields, BaseHiddenFieldsMap):
            hidden_fields = HiddenFieldsCompatibleClass(
                hidden_fields=hidden_fields,
            )
        self._hidden_fields = hidden_fields
        self._convert_underscore = convert_underscore
        self._process_filters = process_filters
        self._model_subclasses = model_subclasses or []
        ResourceMap.add_model_to_resource_mapping(model_class, self)
        for model_subclass in self._model_subclasses:
            ResourceMap.add_model_to_resource_mapping(model_subclass, self)

        self._fields_permissions = (
            fields_permissions
            if fields_permissions is not None
            else field_permissions.UniversalPermissions(
                permission=field_permissions.Permissions.RW
            )
        )

        if not isinstance(self._fields_permissions, field_permissions.BasePermissions):
            raise ValueError(
                "Fields_permissions should inherit"
                "from BasePermissions, not {%s}" % (type(fields_permissions))
            )

    def is_process_filters(self):
        return self._process_filters

    @abc.abstractmethod
    def get_fields(self, override_is_public_field_func=None):
        raise NotImplementedError()

    def get_fields_by_request(self, req):
        """Get fields

        :param req: the webob request
        :return: A dict of fields for specific method
        """

        def is_public_field(model_field_name):
            return self.is_public_field_by_request(
                req=req,
                model_field_name=model_field_name,
            ) and req.api_context.can_be_shown_field(
                self.get_resource_field_name(
                    model_field_name=model_field_name,
                )
            )

        return self.get_fields(override_is_public_field_func=is_public_field)

    def get_fields_by_method(self, method):
        def is_public_field(model_field_name):
            return self.is_public_field_by_method(
                model_field_name=model_field_name,
                method=method,
            )

        return self.get_fields(override_is_public_field_func=is_public_field)

    @abc.abstractmethod
    def get_resource_id(self, model):
        raise NotImplementedError()

    @property
    def _m2r_name_map(self):
        return self._name_map

    @property
    def _r2m_name_map(self):
        return self._inv_name_map

    @property
    def _hidden_model_fields(self):
        return self._hidden_fields

    def get_model_field_name(self, res_field_name):
        name = self._r2m_name_map.get(res_field_name, res_field_name)
        return name.replace("-", "_") if self._convert_underscore else name

    def get_resource_field_name(self, model_field_name):
        name = self._m2r_name_map.get(model_field_name, model_field_name)
        return name.replace("_", "-") if self._convert_underscore else name

    def is_public_field(self, model_field_name):
        return not (
            model_field_name.startswith("_")
            or model_field_name in self._hidden_model_fields
        )

    def is_public_field_by_request(self, req, model_field_name):
        return not (
            model_field_name.startswith("_")
            or self._hidden_fields.is_hidden_field(
                model_field_name=model_field_name,
                req=req,
            )
        )

    def is_public_field_by_method(self, method, model_field_name):
        return not (
            model_field_name.startswith("_")
            or self._hidden_fields.is_hidden_field_by_method(
                model_field_name=model_field_name,
                method=method,
            )
        )

    def get_property_type(self, property_name):
        model = self.get_model()
        return model.properties.properties[property_name].get_property_type()

    def is_id_property(self, property_name):
        model = self.get_model()
        return model.properties[property_name].is_id_property()

    def get_model(self):
        return self._model_class

    def __repr__(self):
        return (
            "<%s[model=%r], name_map=%r, convert_underscore=%s, "
            "process_filters=%s, fields=%r>"
            % (
                self.__class__.__name__,
                self._model_class,
                self._name_map,
                self._convert_underscore,
                self._process_filters,
                self._model_class.properties.properties.keys(),
            )
        )

    def get_prop_kwargs(self, name):
        return self.get_model().properties.properties[name].get_kwargs()

    def generate_schema_object(self, method):
        properties = {}
        required = []
        for name, prop in self.get_fields_by_method(method):
            try:
                prop_kwargs = self.get_prop_kwargs(name)
            except KeyError:
                prop_kwargs = {}
            if prop.is_public():
                properties[prop.api_name] = prop.get_type().to_openapi_spec(prop_kwargs)
                if prop_kwargs.get("required"):
                    required.append(name)
        spec = {
            "type": "object",
            "properties": properties,
        }
        if required:
            spec["required"] = required
        return spec


class ResourceByRAModel(AbstractResource):
    def _prep_field(self, name, prop, override_is_public_field_func=None):
        is_public_field = override_is_public_field_func or self.is_public_field
        if issubclass(prop, ra_properties.BaseProperty):
            return ResourceRAProperty(
                resource=self,
                prop_type=(
                    self._model_class.properties.properties[name].get_property_type()
                ),
                model_property_name=name,
                public=is_public_field(name),
            )
        elif issubclass(prop, ra_relationsips.BaseRelationship):
            return ResourceRelationship(
                self,
                model_property_name=name,
                public=is_public_field(name),
            )
        else:
            raise TypeError("Unknown property type %s" % type(prop))

    def get_field(self, name, override_is_public_field_func=None):
        if not (prop := self._model_class.properties.get(name)):
            raise ValueError("Model doesn't have field %s" % name)
        return self._prep_field(
            name,
            prop,
            override_is_public_field_func,
        )

    def get_fields(self, override_is_public_field_func=None):
        """Get resource fields

        :return: The dict of resource fields.
        """

        for name, prop in self._model_class.properties.items():
            yield name, self._prep_field(name, prop, override_is_public_field_func)

    def get_resource_id(self, model):
        # TODO(efrolov): Write code to convert value to simple value.
        if hasattr(model, "get_id"):
            return str(model.get_id())
        else:
            # TODO(efrolov): Add autosearch resource id by model
            raise ValueError(
                "Can't find resource ID for %s. Please implement "
                "get_id method in your model (%s)" % (model, self._model_class)
            )

    def get_id_type(self):
        id_property = self._model_class.get_id_property()
        if len(id_property) != 1:
            raise TypeError(
                "Model %s returns %s properties which marked as "
                "id_property. Please implement get_id_type "
                "method on your resource %r."
                % (
                    self._model_class,
                    "many" if id_property else "no",
                    type(self),
                )
            )
        return id_property.popitem()[-1].get_property_type()


class ResourceByModelWithCustomProps(ResourceByRAModel):
    def get_field(self, name, override_is_public_field_func=None):
        try:
            return super(ResourceByModelWithCustomProps, self).get_field(
                name=name,
                override_is_public_field_func=override_is_public_field_func,
            )
        except ValueError:
            # native property doesn't exist, try custom property
            pass
        try:
            is_public_field = override_is_public_field_func or self.is_public_field
            return ResourceRAProperty(
                resource=self,
                prop_type=self._model_class.get_custom_property_type(name),
                model_property_name=name,
                public=is_public_field(name),
            )
        except KeyError:
            raise ValueError("Model doesn't have field %s" % name)

    def get_fields(self, override_is_public_field_func=None):
        """Get resource fields

        :return: The dict of resource fields.
        """
        is_public_field = override_is_public_field_func or self.is_public_field

        fields = super(ResourceByModelWithCustomProps, self).get_fields(
            override_is_public_field_func=override_is_public_field_func,
        )

        for name, prop in fields:
            yield name, prop
        for name, prop_type in self._model_class.get_custom_properties():
            yield (
                name,
                ResourceRAProperty(
                    resource=self,
                    prop_type=prop_type,
                    model_property_name=name,
                    public=is_public_field(name),
                ),
            )

    def get_property_type(self, property_name):
        try:
            property_type = super(
                ResourceByModelWithCustomProps,
                self,
            ).get_property_type(property_name=property_name)
        except KeyError:
            model = self.get_model()
            property_type = model.get_custom_property_type(
                property_name=property_name,
            )
        return property_type

    def get_resource_id(self, model):
        return str(model.get_id())



# Copyright 2022 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import mock

from restalchemy.api import packers
from restalchemy.api import controllers

FAKE_LOCATION_PATH = "fake location path"


class TestLocationHeaderLogic(unittest.TestCase):
    def setUp(self):
        super(TestLocationHeaderLogic, self).setUp()
        self._controller = controllers.Controller(None)

    def test_location_for_result(self):
        result = self._controller.process_result("")

        self.assertEqual(result.headers.get("Location", None), None)

    @mock.patch("restalchemy.api.resources.ResourceMap")
    def test_location_for_result_and_add_location(self, resource_map):
        resource_map.get_location.return_value = FAKE_LOCATION_PATH

        result = self._controller.process_result("", add_location=True)

        self.assertEqual(result.headers.get("Location", None), FAKE_LOCATION_PATH)

    def test_location_for_result_and_location_and_tuple_location_false(self):
        result = self._controller.process_result(
            ("", 200, None, False), add_location=True
        )

        self.assertEqual(result.headers.get("Location", None), None)

    @mock.patch("restalchemy.api.resources.ResourceMap")
    def test_location_for_result_and_location_and_tuple_location_true(
        self, resource_map
    ):
        resource_map.get_location.return_value = FAKE_LOCATION_PATH

        result = self._controller.process_result(
            ("", 200, None, True), add_location=True
        )

        self.assertEqual(result.headers.get("Location", None), FAKE_LOCATION_PATH)

    def test_location_for_result_and_tuple_location_false(self):
        result = self._controller.process_result(("", 200, None, False))

        self.assertEqual(result.headers.get("Location", None), None)

    @mock.patch("restalchemy.api.resources.ResourceMap")
    def test_location_for_result_and_tuple_location_true(self, resource_map):
        resource_map.get_location.return_value = FAKE_LOCATION_PATH

        result = self._controller.process_result(("", 200, None, True))

        self.assertEqual(result.headers.get("Location", None), FAKE_LOCATION_PATH)


class BytePacker(packers.JSONPacker):
    def pack(self, obj):
        if isinstance(obj, bytes):
            return obj
        return super().pack(obj)


class ByteController(controllers.Controller):
    __packer__ = BytePacker


class TestRawResponses(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self._controller = ByteController(None)

    def test_binary_result(self):
        headers = {
            "Content-Type": "application/octet-stream",
            "Content-Disposition": 'attachment; filename="test.txt"',
        }

        result = self._controller.process_result((b"1", 200, headers))

        self.assertEqual(result.body, b"1")
        self.assertEqual(result.status, "200 OK")
        self.assertEqual(result.headers["Content-Type"], headers["Content-Type"])
        self.assertEqual(
            result.headers["Content-Disposition"],
            headers["Content-Disposition"],
        )



# Copyright 2024 George Melikov
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.api import packers
from restalchemy.api import resources
from restalchemy.common import exceptions
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.tests.unit import base


class FakeModel(models.ModelWithUUID):
    public = properties.property(types.Integer())
    _hidden = properties.property(types.Integer())


class BasePackerTestCase(base.BaseTestCase):
    def setUp(self):
        super(BasePackerTestCase, self).setUp()
        self._test_instance = packers.BaseResourcePacker(
            resources.ResourceByRAModel(
                FakeModel, hidden_fields=["_hidden"], convert_underscore=False
            ),
            mock.Mock(),
        )

    def tearDown(self):
        super(BasePackerTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_instance

    def test_public_field(self):
        data = {"public": None}

        self._test_instance.unpack(data)

    def test_hidden_field(self):
        data = {"_hidden": None}

        with self.assertRaises(exceptions.ValidationPropertyPrivateError):
            self._test_instance.unpack(data)



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class AbstractClause(metaclass=abc.ABCMeta):
    def __init__(self, value):
        super(AbstractClause, self).__init__()
        self._value = value

    @property
    def value(self):
        return self._value

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.value == other.value

    def __repr__(self):
        return "<%s (%r)>" % (type(self).__name__, self.value)

    def __str__(self):
        return str(self._value)


class EQ(AbstractClause):
    pass


class NE(AbstractClause):
    pass


class GT(AbstractClause):
    pass


class GE(AbstractClause):
    pass


class LT(AbstractClause):
    pass


class LE(AbstractClause):
    pass


class Is(AbstractClause):
    pass


class IsNot(AbstractClause):
    pass


class In(AbstractClause):
    pass


class NotIn(AbstractClause):
    pass


class Like(AbstractClause):
    pass


class NotLike(AbstractClause):
    pass


class AbstractExpression(metaclass=abc.ABCMeta):
    def __init__(self, *clauses):
        super(AbstractExpression, self).__init__()
        self._clauses = clauses

    @property
    def clauses(self):
        return self._clauses

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.clauses == other.clauses

    def __repr__(self):
        return "<%s (%r)>" % (type(self).__name__, self.clauses)


class ClauseList(AbstractExpression):
    pass


class AND(ClauseList):
    pass


class OR(ClauseList):
    pass



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import  # noqa

from mysql.connector import conversion


class MySQLConverter(conversion.MySQLConverter):
    def _list_to_mysql(self, value):
        """Convert list value to mysql driver type

        Convert value to list in for in operation in with statement.
        """
        return [self.to_mysql(item) for item in value]

    def escape(self, value, sql_mode=None):
        """Escape dangerous symbols in value

        Escapes special characters as they are expected to by when MySQL
        receives them.
        As found in MySQL source mysys/charset.c
        Returns the value if not a string, or the escaped string.
        """
        if isinstance(value, list):
            return [self.escape(item, sql_mode) for item in value]
        return super(MySQLConverter, self).escape(value, sql_mode)

    def quote(self, buf):
        """Quote all text values

        Quote the parameters for commands. General rules:
          o numbers are returns as bytes using ascii codec
          o None is returned as bytearray(b'NULL')
          o Everything else is single quoted '<buf>'
        Returns a bytearray object.
        """
        if isinstance(buf, list):
            tmp_list_buf = [self.quote(item) for item in buf]
            tmp_str_buf = b"(%s)" % (b", ".join(tmp_list_buf))
            return bytearray(tmp_str_buf)
        return super(MySQLConverter, self).quote(buf)

    _JSON_to_python = conversion.MySQLConverter._string_to_python

    def _BLOB_to_python(self, value, dsc=None):  # pylint: disable=C0103
        """Convert BLOB data type to Python."""
        if dsc is not None:
            if (
                dsc[7] & conversion.FieldFlag.BLOB
                and dsc[7] & conversion.FieldFlag.BINARY
            ):
                return bytes(value)
        return self._string_to_python(value, dsc)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from urllib.parse import urlparse


import pytest

from restalchemy.tests.functional import consts
from restalchemy.storage.sql import engines


@pytest.fixture(scope="session", autouse=True)
def setup_db_for_worker():
    db_uri = consts.get_database_uri()
    if not (worker_id := os.environ.get("PYTEST_XDIST_WORKER", "")):
        yield
        return

    parsed = urlparse(db_uri)
    worker_db_name_with_slash = f"{parsed.path}_{worker_id}"
    worker_db_name = worker_db_name_with_slash.strip("/")
    db_type = parsed.scheme
    db_created = False
    engines.engine_factory.configure_factory(db_url=db_uri)
    engine = engines.engine_factory.get_engine()
    conn = engine.get_connection()

    # Check if database exists
    if db_type == "postgresql":
        conn.autocommit = True
        c = conn.execute(
            "SELECT 1 FROM pg_database WHERE datname = %s", (worker_db_name,)
        )
        exists = c.fetchall()
    elif db_type == "mysql":
        c = conn.cursor()
        c.execute(
            "SELECT SCHEMA_NAME FROM information_schema.SCHEMATA WHERE SCHEMA_NAME = %s",
            (worker_db_name,),
        )
        exists = c.fetchall()
    else:
        raise ValueError(f"Unsupported database type: {db_type}")

    if not exists:
        db_created = True
        if db_type == "postgresql":
            conn.autocommit = True
            conn.execute(f"CREATE DATABASE {worker_db_name}")
        elif db_type == "mysql":
            conn.cursor().execute(f"CREATE DATABASE {worker_db_name}")

    engine.close_connection(conn)
    del engine
    engines.engine_factory.destroy_engine()

    os.environ["DATABASE_URI"] = parsed._replace(
        path=worker_db_name_with_slash
    ).geturl()
    yield

    if db_created:
        engines.engine_factory.configure_factory(db_url=db_uri)
        engine = engines.engine_factory.get_engine()
        conn = engine.get_connection()
        if db_type == "postgresql":
            conn.autocommit = True
            conn.execute(f"DROP DATABASE IF EXISTS {worker_db_name} WITH (FORCE)")
        elif db_type == "mysql":
            conn.cursor().execute(f"DROP DATABASE IF EXISTS {worker_db_name}")

        engine.close_connection(conn)
        del engine
        engines.engine_factory.destroy_engine()



# Copyright 2023 v.burygin
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from restalchemy.common import singletons


class FakeMetaSingleton(metaclass=singletons.MetaSingleton):
    def __init__(self, prop1, prop2, *args, **kwargs):
        self._prop1 = prop1
        self._prop2 = prop2


class FakeInheritSingleton(singletons.InheritSingleton):
    def __init__(self, prop1, prop2, *args, **kwargs):
        self._prop1 = prop1
        self._prop2 = prop2


class TestSingletons(unittest.TestCase):
    def check_singletons(self, a, b):
        self.assertIs(a, b)
        self.assertTrue(a._prop1)
        self.assertTrue(b._prop1)
        self.assertEqual(a._prop2, ["a", "b", "c"])
        self.assertEqual(b._prop2, ["a", "b", "c"])

    def test_meta(self):
        a = FakeMetaSingleton(prop1=True, prop2=["a", "b", "c"])
        b = FakeMetaSingleton(prop1=False, prop2=["d", "e", "f"])
        self.check_singletons(a, b)

    def test_inherit(self):
        a = FakeInheritSingleton(prop1=True, prop2=["a", "b", "c"])
        b = FakeInheritSingleton(prop1=False, prop2=["d", "e", "f"])
        self.check_singletons(a, b)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import actions
from restalchemy.api import constants
from restalchemy.api import controllers
from restalchemy.api import packers
from restalchemy.api import resources
from restalchemy.common import exceptions as exc
from restalchemy.openapi import constants as oa_c
from restalchemy.openapi import utils
from restalchemy.tests.functional.restapi.ra_based.microservice import (
    storable_models as models,
)


class TagController(controllers.BaseNestedResourceController):
    """Tag controller

    Handle POST http://127.0.0.1:8000/v1/vms/<vm_uuid>/tags/
    Handle GET http://127.0.0.1:8000/v1/vms/<vm_uuid>/tags/
    Handle GET http://127.0.0.1:8000/v1/vms/<vm_uuid>/tags/<tag_name>
    Handle PUT http://127.0.0.1:8000/v1/vms/<vm_uuid>/tags/<tag_name>
    Handle DELETE http://127.0.0.1:8000/v1/vms/<vm_uuid>/tags/<tag_name>
    """

    __resource__ = resources.ResourceByRAModel(models.Tag)
    __pr_name__ = "vm"


class IpAddressController(controllers.BaseNestedResourceController):
    """Port controller

    Handle POST .../v1/vms/<vm_uuid>/ports/<port_uuid>/ip_addresses/
    Handle GET .../v1/vms/<vm_uuid>/ports/<port_uuid>/ip_addresses/
    Handle GET .../v1/vms/<vm_uuid>/ports/<port_uuid>/ip_addresses/<ip_uuid>
    Handle DELETE .../vms/<vm_uuid>/ports/<port_uuid>/ip_addresses/<ip_uuid>
    """

    __resource__ = resources.ResourceByRAModel(models.IpAddress)
    __pr_name__ = "port"


class PortController(controllers.BaseNestedResourceControllerPaginated):
    """Port controller

    Handle POST http://127.0.0.1:8000/v1/vms/<vm_uuid>/ports/
    Handle GET http://127.0.0.1:8000/v1/vms/<vm_uuid>/ports/
    Handle GET http://127.0.0.1:8000/v1/vms/<vm_uuid>/ports/<port_uuid>
    Handle DELETE http://127.0.0.1:8000/v1/vms/<vm_uuid>/ports/<port_uuid>
    """

    __resource__ = resources.ResourceByModelWithCustomProps(
        models.Port,
        process_filters=True,
        hidden_fields=resources.HiddenFieldMap(
            create=["never_call", "some_field1", "unique_field"],
            filter=["never_call", "some_field2"],
            get=["never_call", "some_field3", "unique_field"],
            update=["never_call", "some_field4", "unique_field"],
        ),
    )
    __pr_name__ = "vm"


class PortControllerNone(PortController):
    def get_packer(self, content_type, resource_type=None):
        if content_type == constants.CONTENT_TYPE_APPLICATION_JSON:
            rt = resource_type or self.get_resource()
            return packers.JSONPackerIncludeNullFields(rt, request=self._req)
        return super(PortControllerNone, self).get_packer(content_type, resource_type)


class VMController(controllers.BaseResourceControllerPaginated):
    """VM controller

    Handle POST http://127.0.0.1:8000/v1/vms/
    Handle GET http://127.0.0.1:8000/v1/vms/
    Handle GET http://127.0.0.1:8000/v1/vms/<uuid>
    Handle PUT http://127.0.0.1:8000/v1/vms/<uuid>
    Handle DELETE http://127.0.0.1:8000/v1/vms/<uuid>
    Handle GET http://127.0.0.1:8000/v1/vms/<uuid>/actions/poweron/invoke
    Handle GET http://127.0.0.1:8000/v1/vms/<uuid>/actions/poweroff/invoke
    """

    __resource__ = resources.ResourceByRAModel(models.VM, process_filters=True)

    def create(self, **kwargs):
        """Create VM resource

        API endpoint to create VM resource.

        """
        return super(VMController, self).create(**kwargs)

    @utils.extend_schema(
        summary="Power on virtual machine",
        parameters=[oa_c.build_openapi_parameter("VMUuid")],
        responses=oa_c.build_openapi_get_update_response(
            "{}_{}".format(models.VM.__name__, constants.CREATE.capitalize())
        ),
        tags=["VM"],
    )
    @actions.post
    def poweron(self, resource):
        resource.state = "on"
        resource.save()
        return resource

    @utils.extend_schema(
        summary="Power off virtual machine",
        parameters=[oa_c.build_openapi_parameter("VMUuid")],
        responses=oa_c.build_openapi_get_update_response(
            "{}_{}".format(models.VM.__name__, constants.CREATE.capitalize())
        ),
        tags=["VM"],
    )
    @actions.post
    def poweroff(self, resource, *args, **kwargs):
        resource.state = "off"
        resource.save()
        return resource

    @actions.post
    def power(self, resource, state, *args, **kwargs):
        resource.state = state
        resource.save()
        return resource

    @actions.get
    def power_state(self, resource, *args, **kwargs):
        return {"state": resource.state}


class VMNoProcessFiltersController(VMController):
    __resource__ = resources.ResourceByRAModel(
        models.VMNoProcessFilters,
        process_filters=False,
    )


class VMNoSortController(VMController):
    __resource__ = resources.ResourceByRAModel(models.VMNoSort, process_filters=True)
    __sortable_fields__ = []


class VMDefSortController(VMController):
    __resource__ = resources.ResourceByRAModel(models.VMDefSort, process_filters=True)
    __default_sort__ = {"name": "desc"}


class V1Controller(controllers.RoutesListController):
    __TARGET_PATH__ = "/v1"


class NotImplementedMethodsController(controllers.Controller):
    def filter(self, filters, order_by=None):
        """
        method filter was implemented for testing base error message
        """
        raise exc.NotImplementedError()

    def get(self, uuid):
        """
        method get was implemented for Action testing
        """
        return uuid



#    Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#    Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import dec

from restalchemy.api import contexts
from restalchemy.api import resources
from restalchemy.api import routes


class WSGIApp(object):
    def __init__(self, route_class):
        super(WSGIApp, self).__init__()
        self._main_route = routes.route(route_class)
        resources.ResourceMap.set_resource_map(
            routes.Route.build_resource_map(route_class),
        )

    @property
    def main_route(self):
        return self._main_route

    def process_request(self, req):
        req.api_context = contexts.RequestContext(req)
        return self._main_route(req).do()

    @dec.wsgify
    def __call__(self, req):
        req.application = self
        return self.process_request(req)


class OpenApiApplication(WSGIApp):
    def __init__(self, route_class, openapi_engine):
        super(OpenApiApplication, self).__init__(route_class)
        self._openapi_engine = openapi_engine

    @property
    def openapi_engine(self):
        return self._openapi_engine


Application = WSGIApp



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import copy
import datetime
import uuid
import inspect
import typing as tp

from collections import abc as collections_abc

from restalchemy.common import exceptions as exc
from restalchemy.dm import properties
from restalchemy.dm import types


class DmOperationalStorage(object):
    def __init__(self):
        super(DmOperationalStorage, self).__init__()
        self._storage = {}

    def store(self, name, data):
        self._storage[name] = data

    def get(self, name):
        try:
            return self._storage[name]
        except KeyError:
            raise exc.NotFoundOperationalStorageError(name=name)


class MetaModel(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        props = {}
        attrs["id_properties"] = {}

        for key, value in attrs.copy().items():
            if isinstance(
                value,
                (properties.PropertyCreator, properties.PropertyCollection),
            ):
                props[key] = value
                del attrs[key]
        all_base_properties = properties.PropertyCollection()
        for base in bases:
            base_properties = getattr(base, "properties", None)
            if isinstance(base_properties, properties.PropertyCollection):
                all_base_properties += base_properties
        attrs["properties"] = (
            attrs.pop("properties", properties.PropertyCollection())
            + all_base_properties
            + properties.PropertyCollection(**props)
        )
        for key, prop in attrs["properties"].items():
            if prop.is_id_property():
                attrs["id_properties"][key] = attrs["properties"].properties[key]
        dm_class = super(MetaModel, cls).__new__(cls, name, bases, attrs)
        dm_class.__operational_storage__ = DmOperationalStorage()
        return dm_class

    def __getattr__(cls, name):
        try:
            return cls.properties[name]
        except KeyError:
            raise AttributeError("%s object has no attribute %s" % (cls.__name__, name))

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": "string",
        }
        spec.update(types.build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Model(collections_abc.Mapping, metaclass=MetaModel):
    _python_simple_types = (type(None), str, int, float, complex, bool)

    def __init__(self, **kwargs):
        super(Model, self).__init__()
        self.pour(**kwargs)

    def __getattr__(self, name):
        try:
            return self.properties[name].value
        except KeyError:
            raise AttributeError(
                "%s object has no attribute %s" % (type(self).__name__, name)
            )

    def __setattr__(self, name, value):
        try:
            self.properties[name].value = value
        except KeyError:
            super(Model, self).__setattr__(name, value)
        except exc.TypeError as e:
            raise exc.ModelTypeError(
                property_name=name,
                value=value,
                model=self,
                property_type=e.get_property_type(),
            )
        except exc.ReadOnlyProperty:
            raise exc.ReadOnlyProperty(name=name, model=type(self))

    def pour(self, **kwargs):
        try:
            self.properties = properties.PropertyManager(self.properties, **kwargs)
            self.validate()
        except exc.PropertyRequired as e:
            raise exc.PropertyRequired(name=e.name, model=self.__class__)

        self.id_properties = {}
        for name, prop in self.properties.items():
            if prop.is_id_property():
                self.id_properties[name] = prop

    @classmethod
    def restore(cls, **kwargs):
        obj = cls.__new__(cls)

        # NOTE(aostapenko): We can't invoke 'pour' from __new__ because of
        #                   copy.copy of object becomes imposible
        obj.pour(**kwargs)
        return obj

    def validate(self):
        pass

    def update_dm(self, values):
        for name, value in values.items():
            setattr(self, name, value)

    @classmethod
    def get_id_property(cls):
        if len(cls.id_properties) == 1:
            return cls.id_properties.copy()

        raise TypeError(
            "Model %s has %s properties which marked as "
            "id_property. Please implement get_id_property "
            "method on your model." % (type(cls), "many" if cls.id_properties else "no")
        )

    @classmethod
    def get_id_property_name(cls):
        for key in cls.get_id_property():
            return key

    def get_id_properties(self):
        return self.id_properties.copy()

    def get_data_properties(self):
        result = {}
        for name, prop in self.properties.items():
            if not prop.is_id_property():
                result[name] = prop
        return result

    def is_dirty(self):
        for prop in self.properties.values():
            if prop.is_dirty():
                return True
        return False

    @classmethod
    def get_model_type(cls):
        return cls

    def as_plain_dict(self):
        plain_dict = {}
        props = self.properties

        for name in props:
            val = props[name].value
            if isinstance(val, Model):
                plain_dict[name] = val.get_id()
            elif isinstance(val, self._python_simple_types):
                plain_dict[name] = val
            else:
                plain_dict[name] = copy.deepcopy(val)

        return plain_dict

    def __getitem__(self, name):
        return self.properties[name].value

    def __iter__(self):
        return iter(self.properties)

    def __len__(self):
        return len(self.properties)

    def __str__(self):
        if len(self.id_properties) == 0:
            return self.__repr__()

        return "<%s %s>" % (self.__class__.__name__, self.get_id())

    def __repr__(self):
        result = []
        for k, v in self.items():
            result.append("%s: %s" % (k, v))
        result = ", ".join(result)
        return "<%s {%s}>" % (self.__class__.__name__, result)


class ModelWithID(Model):
    def get_id(self):
        return getattr(self, self.get_id_property_name())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.get_id() == other.get_id()
        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(str(self.get_id()))


class ModelWithUUID(ModelWithID):
    uuid = properties.property(
        types.UUID(),
        read_only=True,
        id_property=True,
        default=lambda: uuid.uuid4(),
    )


class ModelWithRequiredUUID(ModelWithUUID):
    uuid = properties.property(
        types.UUID(), read_only=True, id_property=True, required=True
    )


class CustomPropertiesMixin(object):
    __custom_properties__ = {}

    @classmethod
    def get_custom_properties(cls):
        for name, prop_type in cls.__custom_properties__.items():
            yield name, prop_type

    @classmethod
    def get_custom_property_type(cls, property_name):
        return cls.__custom_properties__[property_name]

    def _check_custom_property_value(self, name, value, static=False, should_be=None):
        prop_type = self.__custom_properties__[name]
        prop_type.validate(value)
        if static and should_be != value:
            raise ValueError(
                ("The value for property `%s` should be `%s` but actual value is `%s`")
                % (name, should_be, value)
            )


class ModelWithTimestamp(Model):
    created_at = properties.property(
        types.UTCDateTimeZ(),
        required=True,
        read_only=True,
        default=lambda: datetime.datetime.now(datetime.timezone.utc),
    )
    updated_at = properties.property(
        types.UTCDateTimeZ(),
        required=True,
        read_only=True,
        default=lambda: datetime.datetime.now(datetime.timezone.utc),
    )

    def update(self, session=None, force=False, *args, **kwargs):
        if self.is_dirty() or force:
            self.properties["updated_at"].set_value_force(
                datetime.datetime.now(datetime.timezone.utc)
            )
        super().update(session=session, force=force, *args, **kwargs)


class ModelWithProject(Model):
    project_id = properties.property(types.UUID(), required=True, read_only=True)


class ModelWithNameDesc(Model):
    name = properties.property(types.String(max_length=255), default="")
    description = properties.property(types.String(max_length=255), default="")


class ModelWithRequiredNameDesc(ModelWithNameDesc):
    name = properties.property(
        types.String(max_length=255),
        required=True,
    )


class DumpToSimpleViewMixin:
    def dump_to_simple_view(
        self,
        skip: tp.Iterable[str] | None = None,
        save_uuid: bool = False,
        custom_properties: bool = False,
    ):
        skip = skip or []
        result = {}
        for name, prop in self.properties.properties.items():
            if name in skip:
                continue
            prop_type = prop.get_property_type()
            if save_uuid and (
                isinstance(prop_type, types.UUID)
                or (
                    isinstance(prop_type, types.AllowNone)
                    and isinstance(prop_type.nested_type, types.UUID)
                )
            ):
                result[name] = getattr(self, name)
                continue

            result[name] = prop_type.to_simple_type(getattr(self, name))

        # Convert the custom properties.
        if not custom_properties and not hasattr(self, "__custom_properties__"):
            return result

        for name, prop_type in self.get_custom_properties():
            result[name] = prop_type.to_simple_type(getattr(self, name))

        return result


class RestoreFromSimpleViewMixin:
    @classmethod
    def restore_from_simple_view(cls, skip_unknown_fields: bool = False, **kwargs):
        model_format = {}
        for name, value in kwargs.items():
            name = name.replace("-", "_")

            # Ignore unknown fields
            if skip_unknown_fields and name not in cls.properties.properties:
                continue

            try:
                prop_type = cls.properties.properties[name].get_property_type()
            except KeyError:
                prop_type = cls.get_custom_property_type(name)
            prop_type = type(prop_type) if not inspect.isclass(prop_type) else prop_type
            if not isinstance(value, prop_type):
                try:
                    model_format[name] = (
                        cls.properties.properties[name]
                        .get_property_type()
                        .from_simple_type(value)
                    )
                except KeyError:
                    model_format[name] = cls.get_custom_property_type(
                        name
                    ).from_simple_type(value)
            else:
                model_format[name] = value
        return cls(**model_format)


class SimpleViewMixin(DumpToSimpleViewMixin, RestoreFromSimpleViewMixin):
    pass
